from abc import ABC
from datetime import datetime, timedelta
from typing import Any, Union, List
import time

# constant
UNLIMITED_MAX_MULTIPLICITY = 9999

class Element(ABC):
    """Element is the Superclass of all structural model elements.

    Args:
        timestamp (datetime): Object creation datetime (default is current time).
        is_derived (bool): Indicates whether the element is derived (False as default).
    
    Attributes:
        timestamp (datetime): Object creation datetime (default is current time).
        is_derived (bool): Indicates whether the element is derived (False as default).
    """

    def __init__(self, timestamp: datetime = None, is_derived: bool = False):
        self.timestamp: datetime = timestamp if timestamp is not None else datetime.now() + \
                         timedelta(microseconds=(time.perf_counter_ns() % 1_000_000) / 1000)
        self.is_derived: bool = is_derived

    @property
    def timestamp(self) -> datetime:
        """str: Get the timestamp of the element."""
        return self.__timestamp

    @timestamp.setter
    def timestamp(self, timestamp: datetime):
        """str: Set the timestamp of the element."""
        self.__timestamp = timestamp

    @property
    def is_derived(self) -> bool:
        """bool: Get whether the element is derived."""
        return self.__is_derived

    @is_derived.setter
    def is_derived(self, is_derived: bool):
        """bool: Set whether the element is derived."""
        self.__is_derived = is_derived

class NamedElement(Element):
    """NamedElement represent a structural element with a name.

    Args:
        name (str): The name of the named element
        timestamp (datetime): Object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the named element (None as default).
        visibility (str): Determines the kind of visibility of the named element (public as default).
        is_derived (bool): Indicates whether the element is derived (False as default).

    Attributes:
        name (str): The name of the named element
        timestamp (datetime): Object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the named element (None as default).
        visibility: Determines the kind of visibility of the named element (public as default).
        is_derived (bool): Indicates whether the element is derived (False as default).
    """

    def __init__(self, name: str, timestamp: datetime = None, synonyms: List[str] = None,
                 visibility: str = "public", is_derived: bool = False):
        super().__init__(timestamp, is_derived)
        self.name: str = name
        self.synonyms: List[str] = synonyms
        self.visibility: str = visibility


    @property
    def name(self) -> str:
        """str: Get the name of the named element."""
        return self.__name

    @name.setter
    def name(self, name: str):
        """
        str: Set the name of the named element.
        
        Raises:
            ValueError: If the name is empty or contains any whitespace characters.
        """
        if ' ' in name:
            raise ValueError(f"'{name}' is invalid. Name cannot contain spaces.")
        self.__name = name

    @property
    def visibility(self) -> str:
        """str: Get the visibility of the named element."""
        return self.__visibility

    @visibility.setter
    def visibility(self, visibility: str):
        """
        str: Set the visibility of the named element.
        
        Raises:
            ValueError: If the visibility provided is none of these: public, 
            private, protected, or package.
        """
        if visibility not in ['public', 'private', 'protected', 'package']:
            raise ValueError("Invalid value of visibility")
        self.__visibility = visibility

    @property
    def synonyms(self) -> List[str]:
        """List[str]: Get the list of synonyms of the named element."""
        return self.__synonyms

    @synonyms.setter
    def synonyms(self, synonyms: List[str]):
        """List[str]: Set the list of synonyms of the named element."""
        self.__synonyms = synonyms

class Type(NamedElement):
    """Type is the Superclass of classes and data types in the model.

    Args:
        name (str): The name of the Type.
        timestamp (datetime): Object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the type (None as default).
        is_derived (bool): Indicates whether the element is derived (False as default).

    Attributes:
        name (str): Inherited from NamedElement, represents the name of the Type.
        timestamp (datetime): Inherited from NamedElement; object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the type (None as default).
        is_derived (bool): Indicates whether the element is derived (False as default).
    """

    def __init__(self, name: str, timestamp: int = None, synonyms: List[str] = None,
                is_derived: bool = False):
        super().__init__(name, timestamp, synonyms, is_derived=is_derived)

    def __repr__(self):
        return f"Type({self.name}, {self.timestamp}, {self.synonyms})"

class DataType(Type):
    """Represents a data type.

    This class inherits from NamedElement and is used to model data types.

    Args:
        name (str): The name of the data type.
        timestamp (datetime): Object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the data type (None as default).

    Attributes:
        name (str): Inherited from NamedElement, represents the name of the data type.
        timestamp (datetime): Inherited from NamedElement; object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the data type (None as default).
    """

    def __repr__(self):
        return f"DataType({self.name})"

class PrimitiveDataType(DataType):
    """Class representing a primitive data type.

    This class is a subclass of DataType and is used to represent primitive data types
    with a specified name.

    Args:
        name (str): The name of the primitive data type.
        timestamp (datetime): Object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the primitive data type (None as default).

    Attributes:
        name (str): Inherited from NamedElement, represents the name of the primitive data type.
        timestamp (datetime): Inherited from NamedElement; object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the primitive data type (None as default).
    """

    @NamedElement.name.setter
    def name(self, name: str):
        """
        str: Set the name of the PrimitiveDataType. 
        
        Raises:
            ValueError: If an invalid primitive data type is provided.
                        Allowed values are int, float, str, bool, time, date, 
                        datetime, timedelta, and any.
        """
        if name not in ['int', 'float', 'str', 'bool', 'time', 'date', 'datetime', 'timedelta']:
            raise ValueError("Invalid primitive data type")
        super(PrimitiveDataType, PrimitiveDataType).name.fset(self, name)

    def __repr__(self):
        return f"PrimitiveDataType({self.name}, {self.timestamp}, {self.synonyms})"

# Define instances of PrimitiveDataType
StringType = PrimitiveDataType("str")
IntegerType = PrimitiveDataType("int")
FloatType = PrimitiveDataType("float")
BooleanType = PrimitiveDataType("bool")
TimeType = PrimitiveDataType("time")
DateType = PrimitiveDataType("date")
DateTimeType = PrimitiveDataType("datetime")
TimeDeltaType = PrimitiveDataType("timedelta")
AnyType = DataType("any")
data_types = {StringType, IntegerType, FloatType, BooleanType,
              TimeType, DateType, DateTimeType, TimeDeltaType, AnyType}

class EnumerationLiteral(NamedElement):
    """Class representing an enumeration literal.

    This class is a subclass of NamedElement and is used to represent individual
    literals within an enumeration.

    Args:
        name (str): The name of the enumeration literal.
        owner (DataType): The owner data type of the enumeration literal (None as default).
        timestamp (datetime): Object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the literal (None as default).

    Attributes:
        name (str): Inherited from NamedElement, represents the name of the enumeration literal.
        owner (DataType): Represents the owner data type of the enumeration literal (None as default).
        timestamp (datetime): Inherited from NamedElement; object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the literal (None as default).
    """

    def __init__(self, name: str, owner: DataType=None, timestamp: int = None, synonyms: List[str] = None):
        super().__init__(name, timestamp, synonyms)
        self.owner: DataType = owner

    @property
    def owner(self) -> DataType:
        """Datatype: Get the owner."""
        return self.__owner

    @owner.setter
    def owner(self, owner: DataType):
        """
        DataType: Set the owner. 
        
        Raises:
            ValueError: If the owner is not an enumeration.
        """
        if isinstance(owner, PrimitiveDataType):
            raise ValueError("Invalid owner")
        self.__owner = owner

    def __repr__(self):
        return f"EnumerationLiteral({self.name}, {self.owner}, {self.timestamp}, {self.synonyms})"

class Enumeration(DataType):
    """Class representing an enumeration.

    This class is a subclass of DataType and is used to represent enumerations
    with a specified name and a set of enumeration literals.

    Args:
        name (str): The name of the enumeration data type.
        literals (set[EnumerationLiteral]): Set of enumeration literals associated with the 
                enumeration (None as default).
        timestamp (datetime): Object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the enumeration (None as default).

    Attributes:
        name (str): Inherited from DataType, represents the name of the enumeration.
        literals (set[EnumerationLiteral]): Represents a set of enumeration literals associated 
                with the enumeration (None as default).
        timestamp (datetime): Inherited from NamedElement; object creation datetime (default is 
                current time).
        synonyms (List[str]): List of synonyms of the enumeration (None as default).
    """

    def __init__(self, name: str, literals: set[EnumerationLiteral] = None, timestamp: int = None,
                 synonyms: List[str] = None):
        super().__init__(name, timestamp, synonyms)
        self.literals: set[EnumerationLiteral] = literals if literals is not None else set()

    @property
    def literals(self) -> set[EnumerationLiteral]:
        """set[EnumerationLiteral]: Get the set of literals."""
        return self.__literals

    @literals.setter
    def literals(self, literals: set[EnumerationLiteral]):
        """
        DataType: Set the literals. 
        
        Raises:
            ValueError: if two literals have the same name.
        """
        if literals is not None:
            names = [literal.name for literal in literals]
            if len(names) != len(set(names)):
                raise ValueError("An enumeration cannot have two literals with the same name")
            for literal in literals:
                literal.owner = self
            self.__literals = literals
        else:
            self.__literals = set()

    def add_literal(self, literal: EnumerationLiteral):
        """
        Add an enumeration literal to the set.
        
        Raises:
            ValueError: if the enumeration literal name already exist.
        """
        if self.literals is not None:
            if literal.name in [literal.name for literal in self.literals]:
                raise ValueError(f"An enumeration cannot have two literals with the same name: '{literal.name}'")
        self.literals.add(literal)

    def __repr__(self):
        return f"Enumeration({self.name}, {self.literals}, {self.timestamp}, {self.synonyms})"

class TypedElement(NamedElement):
    """TypedElement is a subclass of NamedElement and is used to represent elements
    that have a specific type.

    Args:
        name (str): The name of the typed element.
        type (Type, str): The data type of the typed element.
        timestamp (datetime): Object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the typed element (None as default).
        visibility (str): Determines the kind of visibility of the typed element (public as default).
        is_derived (bool): Indicates whether the element is derived (False as default).

    Attributes:
        name (str): Inherited from NamedElement, represents the name of the typed element.
        type (Type): The data type of the typed element.
        timestamp (datetime): Inherited from NamedElement; object creation datetime (default is current time).
        type (Type, str): The data type of the typed element.
        synonyms (List[str]): List of synonyms of the typed element (None as default).
        visibility (str): Inherited from NamedElement, represents the visibility of the typed element (public as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).
    """

    type_mapping = {
        "str": StringType,
        "string": StringType,
        "int": IntegerType,
        "float": FloatType,
        "bool": BooleanType,
        "time": TimeType,
        "date": DateType,
        "datetime": DateTimeType,
        "timedelta": TimeDeltaType
    }

    def __init__(self, name: str, type: Union[Type, str], timestamp: int = None, synonyms: List[str] = None,
                 visibility: str="public", is_derived: bool = False):
        super().__init__(name, timestamp, synonyms, visibility, is_derived)
        self.type = self.type_mapping.get(type, type)

    @property
    def type(self) -> Type:
        """Type: Get the type of the typed element."""
        return self.__type

    @type.setter
    def type(self, type: Type):
        """Type: Set the type of the typed element."""
        self.__type = type

class Multiplicity(Element):
    """Represents the multiplicity of a Property.

    It consists of a minimum and maximum value, indicating the allowed range.

    Args:
        min_multiplicity (int): The minimum multiplicity.
        max_multiplicity (int): The maximum multiplicity. Use "*" for unlimited.
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).

    Attributes:
        min (int): The minimum multiplicity.
        max (int): The maximum multiplicity. Use "*" for unlimited.
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).
    """

    def __init__(self, min_multiplicity: int, max_multiplicity: int, is_derived: bool = False):
        super().__init__(is_derived=is_derived)
        self.min: int = min_multiplicity
        self.max: int = max_multiplicity

    @property
    def min(self) -> int:
        """int: Get the minimum multiplicity."""
        return self.__min

    @min.setter
    def min(self, min_multiplicity: int):
        """
        int: Set the minimum multiplicity 
        
        Raises:
            ValueError: (Invalid min multiplicity) if the minimum multiplicity is less than 0.
        """
        if min_multiplicity < 0:
            raise ValueError("Invalid min multiplicity")
        self.__min = min_multiplicity

    @property
    def max(self) -> int:
        """int: Get the maximum multiplicity."""
        return self.__max

    @max.setter
    def max(self, max_multiplicity: int):
        """
        int: Set the maximum multiplicity.
        
        Raises:
            ValueError: (Invalid max multiplicity) if the maximum multiplicity is less than 0 or
            less than minimum multiplicity.
        """
        if max_multiplicity == "*":
            max_multiplicity = UNLIMITED_MAX_MULTIPLICITY
        if max_multiplicity <= 0:
            raise ValueError("Invalid max multiplicity")
        if max_multiplicity < self.min:
            raise ValueError("Invalid max multiplicity")
        self.__max = max_multiplicity

    def __repr__(self):
        return f'Multiplicity({self.min}, {self.max}, is_derived={self.is_derived})'


# Properties are owned by a class or an association and point to a type with a multiplicity
class Property(TypedElement):
    """A property can represents an attribute of a class or an end of an association.

    Properties are owned by a class or an association.

    Args:
        name (str): The name of the property.
        type (Type): The type of the property.
        owner (Type): The type that owns the property (None as default).
        multiplicity (Multiplicity): The multiplicity of the property (1..1 as default).
        visibility (str): The visibility of the property (public as default).
        is_composite (bool): Indicates whether the property is a composite (False as default).
        is_navigable (bool): Indicates whether the property is navigable in a relationship (True as default).
        is_id (bool): Indicates whether the property is an id (False as default).
        is_read_only (bool): Indicates whether the property is read only (False as default).
        timestamp (datetime): Object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the property (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).

    Attributes:
        name (str): Inherited from TypedElement, represents the name of the property.
        type (Type): Inherited from TypedElement, represents the type of the property.
        owner (Type): The type that owns the property (public as default).
        multiplicity (Multiplicity): The multiplicity of the property (1..1 as default).
        visibility (str): Inherited from TypedElement, represents the visibility of the property (public as default).
        is_composite (bool): Indicates whether the property is a composite (False as default).
        is_navigable (bool): Indicates whether the property is navigable in a relationship (True as default).
        is_id (bool): Indicates whether the property is an id (False as default).
        is_read_only (bool): Indicates whether the property is read only (False as default).
        timestamp (datetime): Inherited from NamedElement; object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the property (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).
    """

    def __init__(self, name: str, type: Type, owner: Type = None, multiplicity: Multiplicity = Multiplicity(1, 1),
                 visibility: str = 'public', is_composite: bool = False, is_navigable: bool = True,
                 is_id: bool = False, is_read_only: bool = False, timestamp: int = None,
                 synonyms: List[str] = None, is_derived: bool = False):
        super().__init__(name, type, timestamp, synonyms, visibility, is_derived)
        self.owner: Type = owner
        self.multiplicity: Multiplicity = multiplicity
        self.is_composite: bool = is_composite
        self.is_navigable: bool = is_navigable
        self.is_id: bool = is_id
        self.is_read_only: bool = is_read_only

    @property
    def owner(self) -> Type:
        """Type: Get the owner type of the property."""
        return self.__owner

    @owner.setter
    def owner(self, owner: Type):
        """
        Type: Set the owner type of the property.
        
        Raises:
            ValueError: (Invalid owner) if the owner is instance of DataType.
        """
        if isinstance(owner, DataType):
            raise ValueError("Invalid owner")
        self.__owner = owner

    @property
    def multiplicity(self) -> Multiplicity:
        """Multiplicity: Get the multiplicity of the property."""
        return self.__multiplicity

    @multiplicity.setter
    def multiplicity(self, multiplicity: Multiplicity):
        """Multiplicity: Set the multiplicity of the property."""
        self.__multiplicity = multiplicity

    @property
    def is_composite(self) -> bool:
        """bool: Get whether the property is composite."""
        return self.__is_composite

    @is_composite.setter
    def is_composite(self, is_composite: bool):
        """bool: Set whether the property is composite."""
        self.__is_composite = is_composite

    @property
    def is_navigable(self) -> bool:
        """bool: Get whether the property is navigable."""
        return self.__is_navigable

    @is_navigable.setter
    def is_navigable(self, is_navigable: bool):
        """bool: Set whether the property is navigable."""
        self.__is_navigable = is_navigable

    @property
    def is_id(self) -> bool:
        """bool: Get whether the property is an id."""
        return self.__is_id

    @is_id.setter
    def is_id(self, is_id: bool):
        """bool: Set whether the property is an id."""
        self.__is_id = is_id

    @property
    def is_read_only(self) -> bool:
        """bool: Get whether the property is read only."""
        return self.__is_read_only

    @is_read_only.setter
    def is_read_only(self, is_read_only: bool):
        """bool: Set whether the property is read only."""
        self.__is_read_only = is_read_only

    def __repr__(self):
        return (
            f'Property({self.name}, {self.visibility}, {self.type}, {self.multiplicity}, '
            f'is_composite={self.is_composite}, is_id={self.is_id}, '
            f'is_read_only={self.is_read_only}, {self.timestamp}, {self.synonyms}, '
            f'is_derived={self.is_derived})'
        )

class Parameter(TypedElement):
    """
    Parameter is used to represent a parameter of a method with a specific type.

    Args:
        name (str): The name of the parameter.
        type (Type): The data type of the parameter.
        default_value (Any): The default value of the parameter (None as default).
        timestamp (datetime): Object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the parameter (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).

    Attributes:
        name (str): Inherited from NamedElement, represents the name of the parameter.
        type (Type): Inherited from TypedElement, represents the type of the parameter.
        default_value (Any): The default value of the parameter (None as default).
        timestamp (datetime): Inherited from NamedElement; object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the parameter (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).
    """

    def __init__(self, name: str, type: Type, default_value: Any = None, timestamp: int = None,
                 synonyms: List[str] = None, is_derived: bool = False):
        super().__init__(name, type, timestamp, synonyms, is_derived=is_derived)
        self.default_value: Any = default_value

    @property
    def default_value(self) -> Any:
        """Type: Get the default value of the parameter."""
        return self.__default_value

    @default_value.setter
    def default_value(self, default_value: Any):
        """Type: Set the default value of the parameter."""
        self.__default_value = default_value

    def __repr__(self):
        return (
            f'Parameter({self.name}, {self.type}, {self.default_value}, {self.timestamp}, '
            f'{self.synonyms}, is_derived={self.is_derived})'
        )

class Method(TypedElement):
    """
    Method is used to represent a method of a class.

    Args:
        name (str): The name of the method.
        visibility (str): Determines the kind of visibility of the method (public as default).
        is_abstract (bool): Indicates if the method is abstract (False as default).
        parameters (set[Parameter]): The set of parameters for the method (set() as default).
        type (Type): The type of the method (None as default).
        owner (Type): The type that owns the method (None as default).
        code (str): code of the method ("" as default).
        timestamp (datetime): Object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the method (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).

    Attributes:
        name (str): Inherited from TypedElement, represents the name of the method.
        visibility (str): Inherited from TypedElement, represents the visibility of the method (public as default).
        is_abstract (bool): Indicates if the method is abstract. (False as default)
        parameters (set[Parameter]): The set of parameters for the method (set() as default).
        type (Type): Inherited from TypedElement, represents the type of the method (None as default).
        owner (Type): The type that owns the property (None as default).
        code (str): code of the method ("" as default).
        timestamp (datetime): Inherited from NamedElement; object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the method (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).
    """

    def __init__(self, name: str, visibility: str = "public", is_abstract: bool = False,
                 parameters: set[Parameter] = None, type: Type = None, owner: Type = None,
                 code: str = "", timestamp: int = None, synonyms: List[str] = None, is_derived: bool = False):
        super().__init__(name, type, timestamp, synonyms, visibility, is_derived)
        self.is_abstract: bool = is_abstract
        self.parameters: set[Parameter] = parameters if parameters is not None else set()
        self.owner: Type = owner
        self.code: str = code

    @property
    def is_abstract(self) -> bool:
        """bool: Get whether the method is abstract."""
        return self.__is_abstract

    @is_abstract.setter
    def is_abstract(self, is_abstract: bool):
        """bool: Set whether the method is abstract."""
        self.__is_abstract = is_abstract

    @property
    def parameters(self) -> set[Parameter]:
        """set[Parameter]: Get the set of parameters of the method."""
        return self.__parameters

    @parameters.setter
    def parameters(self, parameters: set[Parameter]):
        """
        set[Parameter]: Set the parameters of the method.
        
        Raises:
            ValueError: if two parameters have the same name.
        """
        if parameters is not None:
            names_seen = set()
            duplicates = set()

            for parameter in parameters:
                if parameter.name in names_seen:
                    duplicates.add(parameter.name)
                names_seen.add(parameter.name)
                parameter.owner = self

            if duplicates:
                raise ValueError(f"A method cannot have parameters with duplicate names: {', '.join(duplicates)}")

            self.__parameters = parameters
        else:
            self.__parameters = set()

    def add_parameter(self, parameter: Parameter):
        """
        Parameter: Add a parameter to the set of class parameters.
        
        Raises:
            ValueError: if the parameter name already exist.
        """
        if self.parameters is not None:
            if parameter.name in [parameter.name for parameter in self.parameters]:
                raise ValueError(f"A method cannot have two parameters with the same name: '{parameter.name}'")
        self.parameters.add(parameter)

    @property
    def owner(self) -> Type:
        """Type: Get the owner type of the method."""
        return self.__owner

    @owner.setter
    def owner(self, owner: Type):
        """
        Type: Set the owner type of the method.
        
        Raises:
            ValueError: (Invalid owner) if the owner is instance of DataType.
        """
        if isinstance(owner, DataType):
            raise ValueError("Invalid owner")
        self.__owner = owner

    @property
    def code(self) -> str:
        """str: Get the code of the method."""
        return self.__code

    @code.setter
    def code(self, code: str):
        """str: Set the code of the method."""
        self.__code = code

    def __repr__(self):
        return (
            f'Method({self.name}, {self.visibility}, is_abstract={self.is_abstract}, {self.parameters}, '
            f'{self.type}, {self.owner}, {self.code}, {self.timestamp}, {self.synonyms}, '
            f'is_derived={self.is_derived})'
        )


class BehaviorImplementation(NamedElement):
    """A behaviorImplementation represents the body of a behavior associated with a class.

    Args:
        name (str): The name of the behavior implementation.
        
    Attributes:
        name (str): The name of the behavior implementation.  
    """
    
    def __init__(self, name: str):
        super().__init__(name)


    def __repr__(self):
        return f'BehaviorImplementation({self.name})'


class BehaviorDeclaration(NamedElement):
    """A BehaviorDeclaration represents the signature of a behavior associated with a class.

    Args:
        name (str): The name of the behavior.
        implementations (set[BehaviorImplementation]): The implementations associated with the behavior.
        
    Attributes:
        name (str): The name of the behavior.
        implementations (set[BehaviorImplementation]): The implementations associated with the behavior.
    """
    
    def __init__(self, name: str, implementations: set[BehaviorImplementation]):
        super().__init__(name)
        self.implementations: set[BehaviorImplementation] = implementations


    @property
    def implementations(self) -> set[BehaviorImplementation]:
        """set[BehaviorImplementation]: Get the implementations of the behavior."""
        return self.__implementations


    @implementations.setter
    def implementations(self, implementations: set[BehaviorImplementation]):
        """
        set[BehaviorImplementation]: Set the implementations of the behavior.

        Raises:
            ValueError: if two implementations have the same name.
        """
        if implementations is not None:
            names = [implementation.name for implementation in implementations]
            if len(names) != len(set(names)):
                raise ValueError("A behavior cannot have two implementations with the same name")
            self.__implementations = implementations
        else:
            self.__implementations = set()

    def __repr__(self):
        return f'BehaviorDeclaration({self.name}, {self.implementations})'


class Class(Type):
    """Represents a class in a modeling context.

    A Class is a type that defines a blueprint for objects. It can have attributes, associations,
    and generalizations with other classes.

    Args:
        name (str): The name of the class.
        attributes (set[Property]): The set of attributes associated with the class.
        behaviors (set[BehaviorDeclaration]): The set of behaviors associated with the class (None as default).
        is_abstract (bool): Indicates whether the class is abstract.
        is_read_only (bool): Indicates whether the class is read only.

    Attributes:
        name (str): Inherited from Type, represents the name of the class.
        attributes (set[Property]): The set of attributes associated with the class.
        behaviors (set[BehaviorDeclaration]): The set of behaviors associated with the class (None as default).
        is_abstract (bool): Indicates whether the class is abstract.
        is_read_only (bool): Indicates whether the class is read only.
        attributes (set[Property]): The set of attributes associated with the class (set() as default).
        methods (set[Method]): The set of methods of the class (set() as default).
        is_abstract (bool): Indicates whether the class is abstract (False as default).
        is_read_only (bool): Indicates whether the class is read only (False as default).
        behaviors (set[BehaviorDeclaration]): The set of behaviors associated with the class (None as default).
        timestamp (datetime): Object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the class (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).

    Attributes:
        name (str): Inherited from Type, represents the name of the class.
        attributes (set[Property]): The set of attributes associated with the class (set() as default).
        methods (set[Method]): The set of methods of the class (set() as default).
        is_abstract (bool): Indicates whether the class is abstract (False as default).
        is_read_only (bool): Indicates whether the class is read only (False as default).
        behaviors (set[BehaviorDeclaration]): The set of behaviors associated with the class (None as default).
        __associations (set[Association]): Set of associations involving the class.
        __generalizations (set[Generalization]): Set of generalizations involving the class.
        timestamp (datetime): Inherited from NamedElement; object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the class (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).
    """

    def __init__(self, name: str, attributes: set[Property] = None, methods: set[Method] = None,
                 is_abstract: bool= False, is_read_only: bool= False, behaviors: set[BehaviorDeclaration] = None,
                 timestamp: int = None, synonyms: List[str] = None, is_derived: bool = False):
        super().__init__(name, timestamp, synonyms, is_derived=is_derived)
        self.is_abstract: bool = is_abstract
        self.is_read_only: bool = is_read_only
        self.behaviors: set[BehaviorDeclaration] = behaviors if behaviors is not None else set()
        self.attributes: set[Property] = attributes if attributes is not None else set()
        self.methods: set[Method] = methods if methods is not None else set()
        self.__associations: set[Association] = set()
        self.__generalizations: set[Generalization] = set()

    @property
    def attributes(self) -> set[Property]:
        """set[Property]: Get the attributes of the class."""
        return self.__attributes

    @attributes.setter
    def attributes(self, attributes: set[Property]):
        """
        set[Property]: Set the attributes of the class.
        
        Raises:
            ValueError: if two attributes have the same name.
            ValueError: if two attributes are id.
        """
        if attributes is not None:
            names_seen = set()
            duplicates = set()
            id_counter = 0

            for attribute in attributes:
                if attribute.name in names_seen:
                    duplicates.add(attribute.name)
                names_seen.add(attribute.name)

                if attribute.is_id:
                    id_counter += 1

            if duplicates:
                raise ValueError(f"A class cannot have attributes with duplicate names: {', '.join(duplicates)}")

            if id_counter > 1:
                raise ValueError("A class cannot have more than one attribute marked as 'id'")

            for attribute in attributes:
                if attribute.owner and attribute.owner != self:
                    attribute.owner.attributes.discard(attribute)
                attribute.owner = self
            self.__attributes = attributes
        else:
            self.__attributes = set()

    @property
    def methods(self) -> set[Method]:
        """set[Method]: Get the methods of the class."""
        return self.__methods

    @methods.setter
    def methods(self, methods: set[Method]):
        """
        set[Method]: Set the methods of the class.
        
        Raises:
            ValueError: if two methods have the same name.
        """
        if methods is not None:
            names_seen = set()
            duplicates = set()

            for method in methods:
                if method.name in names_seen:
                    duplicates.add(method.name)
                names_seen.add(method.name)

            if duplicates:
                raise ValueError(f"A class cannot have methods with duplicate names: {', '.join(duplicates)}")

            for method in methods:
                method.owner = self
            self.__methods = methods
        else:
            self.__methods = set()

    def add_method(self, method: Method):
        """
        Method: Add a method to the set of class methods.
        
        Raises:
            ValueError: if the method name already exist.
        """
        if self.methods is not None:
            if method.name in [method.name for method in self.methods]:
                raise ValueError(f"A class cannot have two methods with the same name: '{method.name}'")
        method.owner = self
        self.methods.add(method)

    def all_attributes(self) -> set[Property]:
        """set[Property]: Get all attributes, including inherited ones."""
        inherited_attributes: set[Property] = self.inherited_attributes()
        return self.__attributes | inherited_attributes

    def add_attribute(self, attribute: Property):
        """
        Property: Add an attribute to the set of class attributes.
        
        Raises:
            ValueError: if the attribute name already exist.
        """
        if self.attributes is not None:
            if attribute.name in [attribute.name for attribute in self.attributes]:
                raise ValueError(f"A class cannot have two attributes with the same name: '{attribute.name}'")
        if attribute.owner and attribute.owner != self:
            attribute.owner.attributes.discard(attribute)
        attribute.owner = self
        self.attributes.add(attribute)

    @property
    def behaviors(self) -> set[BehaviorDeclaration]:
        """set[BehaviorDeclaration]: Get the behaviors associated with the class."""
        return self.__behaviors

    @behaviors.setter
    def behaviors(self, behaviors: set[BehaviorDeclaration]):
        """
        set[BehaviorDeclaration]: Set the behaviors associated with the class.
        
        Raises:
            ValueError: if two behaviors have the same name.
        """
        if behaviors is not None:
            names = [behavior.name for behavior in behaviors]
            if len(names) != len(set(names)):
                raise ValueError("A class cannot have two behaviors with the same name")
            self.__behaviors = behaviors
        else:
            self.__behaviors = set()

    @property
    def is_abstract(self) -> bool:
        """bool: Get whether the class is abstract."""
        return self.__is_abstract

    @is_abstract.setter
    def is_abstract(self, is_abstract: bool):
        """bool: Set whether the class is abstract."""
        self.__is_abstract = is_abstract

    @property
    def is_read_only(self) -> bool:
        """bool: Get whether the class is read only."""
        return self.__is_read_only

    @is_read_only.setter
    def is_read_only(self, is_read_only: bool):
        """bool: Set whether the class is read only."""
        self.__is_read_only = is_read_only

    @property
    def associations(self) -> set:
        """set[Association]: Get the set of associations involving the class."""
        return self.__associations

    def _add_association(self, association):
        """Association: Add an association to the set of class associations."""
        self.__associations.add(association)

    def _delete_association(self, association):
        """Association: Remove an association to the set of class associations."""
        self.__associations.discard(association)

    @property
    def generalizations(self) -> set:
        """set[Generalization]: Get the set of generalizations involving the class."""
        return self.__generalizations

    def _add_generalization(self, generalization):
        """Generalization: Add a generalization to the set of class generalizations."""
        self.__generalizations.add(generalization)

    def _delete_generalization(self, generalization):
        """Generalization: Remove a generalization to the set of class generalizations."""
        self.__generalizations.discard(generalization)

    def inherited_attributes(self) -> set[Property]:
        """set[Property]: Get the set of inherited attributes."""
        inherited_attributes = set()
        for parent in self.all_parents():
            inherited_attributes.update(parent.attributes)
        return inherited_attributes

    def association_ends(self) -> set:
        """set[Property]: Get the set of association ends of the class."""
        ends = set()
        for association in self.__associations:
            aends = association.ends
            ends.update(aends)
            l_aends = list(aends)
            if not(len(l_aends) == 2 and l_aends[0].type == l_aends[1].type):
                for end in aends:
                    if end.type == self:
                        ends.discard(end)
        return ends

    def all_association_ends(self) -> set[Property]:
        """set[Property]: Get the set of direct and indirect association ends of the class."""
        all_ends = self.association_ends()
        for parent in self.all_parents():
            ends = parent.association_ends()
            all_ends.update(ends)
        return all_ends

    def parents(self) -> set:
        """set[Class]: Get the set of direct parents of the class."""
        parents = set()
        for generalization in self.__generalizations:
            if generalization.general != self:
                parents.add(generalization.general)
        return parents

    def all_parents(self) -> set:
        """set[Class]: Get the set of direct and indirect parents of the class."""
        all_parents = set()
        all_parents.update(self.parents())
        for parent in self.parents():
            all_parents.update(parent.all_parents())
        return all_parents

    def specializations(self) -> set:
        """set[Class]: Get the set of direct specializations (children) of the class."""
        specializations = set()
        for generalization in self.__generalizations:
            if generalization.specific != self:
                specializations.add(generalization.specific)
        return specializations

    def all_specializations(self) -> set:
        """set[Class]: Get the set of direct and indirect specializations (children) of the class."""
        all_spec = set()
        all_spec.update(self.specializations())
        for specialization in self.specializations():
            all_spec.update(specialization.all_specializations())
        return all_spec

    def id_attribute(self) -> Property:
        """Property: Get the id attribute of the class."""
        for attribute in self.attributes:
            if attribute.is_id:
                return attribute
        return None

    def __repr__(self):
        return (
                f'Class({self.name}, {self.attributes}, {self.methods}, {self.timestamp}, {self.synonyms}, '
                f'is_abstract={self.is_abstract}, is_derived={self.is_derived})'
        )


class Association(NamedElement):
    """Represents an association between classes.

    An Association defines a relationship between classes and is composed of two or more ends,
    each associated with a class. An association must have more than one end.

    Args:
        name (str): The name of the association.
        ends (set[Property]): The set of ends related to the association.
        timestamp (datetime): Object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the association (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).
        
    Attributes:
        name (str): Inherited from NamedElement, represents the name of the association.
        ends (set[Property]): The set of ends related to the association.
        timestamp (datetime): Inherited from NamedElement; object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the association (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).
    """

    def __init__(self, name: str, ends: set[Property], timestamp: int = None, synonyms: List[str] = None,
                is_derived: bool = False):
        super().__init__(name, timestamp, synonyms, is_derived=is_derived)
        self.ends: set[Property] = ends

    @property
    def ends(self) -> set[Property]:
        """set[Property]: Get the ends of the association."""
        return self.__ends

    @ends.setter
    def ends(self, ends: set[Property]):
        """
        set[Property]: Set the ends of the association. Two or more ends are required.
        
        Raises:
            ValueError: if an association has less than two ends.
        """
        if len(ends) <= 1:
            raise ValueError("An association must have more than one end")
        if hasattr(self, "ends"):
            for end in self.ends:
                end.type._delete_association(association=self)
        for end in ends:
            end.owner = self
            end.type._add_association(association=self)
        self.__ends = ends

    def __repr__(self):
        return f'Association({self.name}, {self.ends}, {self.timestamp}, {self.synonyms}, is_derived={self.is_derived})'

class BinaryAssociation(Association):
    """Represents a binary association between two classes.

    A BinaryAssociation is a specialized form of Association that specifically involves
    two ends, each associated with a class. It enforces constraints on the association,
    such as having exactly two ends. Exactly two ends are required 

    Args:
        name (str): The name of the binary association.
        ends (set[Property]): The set of ends related to the binary association.
        timestamp (datetime): Object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the binary association (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).

    Attributes:
        name (str): Inherited from Association, represents the name of the binary association.
        ends (set[Property]): Inherited from NamedElement, represents the set of ends related to the binary association.
        timestamp (datetime): Inherited from NamedElement; object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the binary association (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).
    """

    @Association.ends.setter
    def ends(self, ends: set[Property]):
        """set[Property]: Set the ends of the association.
        
        Raises:
            ValueError: if the associaiton ends are not exactly two, or if both ends are tagged as agregation, or 
            if both ends are tagged as composition.
        """
        if len(ends) != 2:
            raise ValueError("A binary association must have exactly two ends")
        if list(ends)[0].is_composite is True and list(ends)[1].is_composite is True:
            raise ValueError("The composition attribute cannot be tagged at both ends")
        super(BinaryAssociation, BinaryAssociation).ends.fset(self, ends)

    def __repr__(self):
        return f'BinaryAssociation({self.name}, {self.ends}, {self.timestamp}, {self.synonyms}, is_derived={self.is_derived})'

class AssociationClass(Class):
    # Class that has an association nature
    """An AssociationClass is a class that that has an association nature.
    It inherits from Class and is associated with an underlying Association.

    Args:
        name (str): The name of the association class.
        attributes (set[Property]): The set of attributes associated with the association class.
        association (Association): The underlying association linked to the association class.
        timestamp (datetime): Object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the association class (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).

    Attributes:
        name (str): Inherited from Class, represents the name of the association class.
        attributes (set[Property]): Inherited from Class, represents the set of attributes associated with the association class.
        association (Association): The underlying association linked to the association class.
        timestamp (datetime): Inherited from NamedElement; object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the association class (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).
    """

    def __init__(self, name: str, attributes: set[Property], association: Association, timestamp: int = None,
                 synonyms: List[str] = None, is_derived: bool = False):
        super().__init__(name, attributes, timestamp, synonyms, is_derived=is_derived)
        self.association: Association = association

    @property
    def association(self) -> Association:
        """Association: Get the underlying association of the association class."""
        return self.__association

    @association.setter
    def association(self, association: Association):
        """Association: Set the underlying association of the association class."""
        self.__association = association

    def __repr__(self):
        return (
            f'AssociationClass({self.name}, {self.attributes}, {self.association}, {self.timestamp}, '
            f'{self.synonyms}, is_derived={self.is_derived})'
        )

class Generalization(Element):
    """Represents a generalization relationship between two classes.

    A Generalization is a relationship between two classes, where one class (specific)
    inherits attributes and behaviors from another class (general).

    Args:
        general (Class): The general (parent) class in the generalization relationship.
        specific (Class): The specific (child) class in the generalization relationship.
        timestamp (datetime): Object creation datetime (default is current time).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).
    
    Attributes:
        general (Class): The general (parent) class in the generalization relationship.
        specific (Class): The specific (child) class in the generalization relationship.
        timestamp (datetime): Inherited from NamedElement; object creation datetime (default is current time).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).
    """

    def __init__(self, general: Class, specific: Class, timestamp: int = None, is_derived: bool = False):
        super().__init__(timestamp, is_derived)
        self.general: Class = general
        self.specific: Class = specific

    @property
    def general(self) -> Class:
        """Class: Get the general (parent) class."""
        return self.__general

    @general.setter
    def general(self, general: Class):
        """Class: Set the general (parent) class."""
        if hasattr(self, "general"):
            self.general._delete_generalization(generalization=self)
        general._add_generalization(generalization=self)
        self.__general = general

    @property
    def specific(self) -> Class:
        """Class: Get the specific (child) class."""
        return self.__specific

    @specific.setter
    def specific(self, specific: Class):
        """
        Class: Set the specific (child) class.
        
        Raises:
            ValueError: if the general class is equal to the specific class
        """
        if specific == self.general:
            raise ValueError("A class cannot be a generalization of itself")
        if hasattr(self, "specific"):
            self.specific._delete_generalization(generalization=self)
        specific._add_generalization(generalization=self)
        self.__specific = specific

    def __repr__(self):
        return f'Generalization({self.general}, {self.specific}, {self.timestamp}, is_derived={self.is_derived})'

class GeneralizationSet(NamedElement):
    """Represents a set of generalization relationships.

    Args:
        name (str): The name of the generalization set.
        generalizations (set[Generalization]): The set of generalization relationships in the set.
        is_disjoint (bool): Indicates whether the set is disjoint (instances cannot belong to more than one class
            in the set).
        is_complete (bool): Indicates whether the set is complete (every instance of the superclass must belong to
            a subclass).
        timestamp (datetime): Object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the generalization set (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).

    Attributes:
        name (str): Inherited from NamedElement, represents the name of the generalization set.
        generalizations (set[Generalization]): The set of generalization relationships in the set.
        is_disjoint (bool): Indicates whether the set is disjoint (instances cannot belong to more than one class
            in the set).
        is_complete (bool): Indicates whether the set is complete (every instance of the superclass must belong to
            a subclass).
        timestamp (datetime): Inherited from NamedElement; object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the generalization set (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).
    """

    def __init__(self, name: str, generalizations: set[Generalization], is_disjoint: bool, is_complete: bool,
                timestamp: int = None, synonyms: List[str] = None, is_derived: bool = False):
        super().__init__(name, timestamp, synonyms, is_derived=is_derived)
        self.generalizations: set[Generalization] = generalizations
        self.is_disjoint: bool = is_disjoint
        self.is_complete: bool = is_complete

    @property
    def generalizations(self) -> set[Generalization]:
        """set[Generalization]: Get the generalization relationships."""
        return self.__generalizations

    @generalizations.setter
    def generalizations(self, generalizations: set[Generalization]):
        """set[Generalization]: Set the generalization relationships."""
        self.__generalizations = generalizations

    @property
    def is_disjoint(self) -> bool:
        """bool: Get whether the set is disjoint."""
        return self.__is_disjoint

    @is_disjoint.setter
    def is_disjoint(self, is_disjoint: bool):
        """bool: Set whether the set is disjoint."""
        self.__is_disjoint = is_disjoint

    @property
    def is_complete(self) -> bool:
        """bool: Get whether the set is complete."""
        return self.__is_complete

    @is_complete.setter
    def is_complete(self, is_complete: bool):
        """bool: Set whether the set is complete."""
        self.__is_complete = is_complete

    def __repr__(self):
        return (
            f'GeneralizationSet({self.name}, {self.generalizations}, '
            f'is_disjoint={self.is_disjoint}, is_complete={self.is_complete}, {self.timestamp}, '
            f'{self.synonyms}, is_derived={self.is_derived})'
        )

class Package(NamedElement):
    """A Package is a grouping mechanism that allows organizing and managing a set of NamedElements.

    Attributes:
        name (str): The name of the package.
        elements (set[NamedElement]): The set of elements contained in the package.
        timestamp (datetime): Object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the package (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).
    
    Attributes:
        name (str): Inherited from NamedElement, represents the name of the package.
        elements (set[NamedElement]): The set of elements contained in the package.
        timestamp (datetime): Inherited from NamedElement; object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the package (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).
    """

    def __init__(self, name: str, elements: set[NamedElement], timestamp: int = None, synonyms: List[str] = None,
                is_derived: bool = False):
        super().__init__(name, timestamp, synonyms, is_derived=is_derived)
        self.elements: set[NamedElement] = elements

    @property
    def elements(self) -> set[NamedElement]:
        """set[NamedElement]: Get the named elements contained in the package."""
        return self.__elements

    @elements.setter
    def elements(self, elements: set[NamedElement]):
        """set[NamedElement]: Set the named elements contained in the package."""
        self.__elements = elements

    def get_classes(self) -> set[Class]:
        """set[Class]: Get all classes within the package."""
        return {element for element in self.elements if isinstance(element, Class)}

    def get_associations(self) -> set[Association]:
        """set[Association]: Get all associations within the package."""
        return {element for element in self.elements if isinstance(element, Association)}

    def get_generalizations(self) -> set[Generalization]:
        """set[Generalization]: Get all generalizations within the package."""
        return {element for element in self.elements if isinstance(element, Generalization)}

    def get_enumerations(self) -> set[Enumeration]:
        """set[Enumeration]: Get all enumerations within the package."""
        return {element for element in self.elements if isinstance(element, Enumeration)}

    def __repr__(self):
        return f'Package({self.name}, {self.elements}), {self.timestamp}, {self.synonyms}, is_derived={self.is_derived}'

class Constraint(NamedElement):
    """A Constraint is a statement that restricts or defines conditions on the behavior,
    structure, or other aspects of the modeled system.

    Args:
        name (str): The name of the constraint.
        context (Class): The class to which the constraint is associated.
        expression (str): The expression or condition defined by the constraint.
        language (str): The language in which the constraint expression is written.
        timestamp (datetime): Object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the constraint (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).

    Attributes:
        name (str): Inherited from NamedElement, represents the name of the constraint.
        context (Class): The class to which the constraint is associated.
        expression (str): The expression or condition defined by the constraint.
        language (str): The language in which the constraint expression is written.
        timestamp (datetime): Inherited from NamedElement; object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the constraint (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).
    """

    def __init__(self, name: str, context: Class, expression: Any, language: str, timestamp: int = None,
                 synonyms: List[str] = None, is_derived: bool = False):
        super().__init__(name, timestamp, synonyms, is_derived=is_derived)
        self.context: Class = context
        self.expression: str = expression
        self.language: str = language

    @property
    def context(self) -> Class:
        """Class: Get the class to which the constraint is associated."""
        return self.__context

    @context.setter
    def context(self, context: Class):
        """Class: Set the class to which the constraint is associated."""
        self.__context = context

    @property
    def expression(self) -> str:
        """str: Get the expression or condition defined by the constraint."""
        return self.__expression

    @expression.setter
    def expression(self, expression: Any):
        """str: Set the expression or condition defined by the constraint."""
        self.__expression = expression

    @property
    def language(self) -> str:
        """str: Get the language in which the constraint expression is written."""
        return self.__language

    @language.setter
    def language(self, language: str):
        """str: Set the language in which the constraint expression is written."""
        self.__language = language

    def __repr__(self):
        return (
            f'Constraint({self.name}, {self.context.name}, {self.language}, {self.expression}, '
            f'{self.timestamp}, is_derived={self.is_derived})'
        )

class Model(NamedElement):
    """A model is the root element. A model is the root element. There are different types of models
    that inherit from this class. For example, DomainModel, ObjectModel, or GUIModel.

    Args:
        name (str): The name of the model.
        timestamp (datetime): Object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the model (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).
        elements (set[Element]): Set of model Elements in the Model.
        
    Attributes:
        name (str): Inherited from NamedElement, represents the name of the model.
        timestamp (datetime): Inherited from NamedElement; object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the model (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).
        elements (set[Element]): Set of model Elements in the Model.
    """
    def __init__(self, name: str, timestamp: int = None, synonyms: List[str] = None, is_derived: bool = False,
                elements: set[Element] = None):
        super().__init__(name, timestamp, synonyms, is_derived=is_derived)
        self.elements: set[Element] = elements if elements is not None else set()

    @property
    def elements(self) -> set[Element]:
        """set[Element]: Get the set of model elements in the model."""
        return self.__elements

    @elements.setter
    def elements(self, elements: set[Element]):
        """set[Element]: Set the set of model elements in the model."""
        if elements is not None:
            self.__elements = elements
        else:
            self.__elements = set()

class DomainModel(Model):
    """A domain model comprises a number of types, associations, 
    generalizations, packages, constraints, and others.

    Args:
        name (str): The name of the domain model.
        types (set[Type]): The set of types (classes and datatypes) in the domain model (set() as default).
        associations (set[Association]): The set of associations in the domain model (set() as default).
        generalizations (set[Generalization]): The set of generalizations in the domain model (set() as default).
        packages (set[Package]): The set of packages in the domain model (set() as default).
        constraints (set[Constraint]): The set of constraints in the domain model (set() as default).
        timestamp (datetime): Object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the domain model (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).
        elements (set[Element]): Set of model Elements in the Model. This property is derived (auto-calculated).

    Attributes:
        name (str): Inherited from NamedElement, represents the name of the domain model.
        types (set[Type]): The set of types (classes and datatypes) in the domain model (set() as default).
        associations (set[Association]): The set of associations in the domain model (set() as default).
        generalizations (set[Generalization]): The set of generalizations in the domain model (set() as default).
        packages (set[Package]): The set of packages in the domain model (set() as default).
        constraints (set[Constraint]): The set of constraints in the domain model (set() as default).
        timestamp (datetime): Inherited from NamedElement; object creation datetime (default is current time).
        synonyms (List[str]): List of synonyms of the domain model (None as default).
        is_derived (bool): Inherited from NamedElement, indicates whether the element is derived (False as default).
        elements (set[Element]): Set of model Elements in the Model. This property is derived (auto-calculated).
    """

    def __init__(self, name: str, types: set[Type] = None, associations: set[Association] = None,
                generalizations: set[Generalization] = None, packages: set[Package] = None,
                constraints: set[Constraint] = None, timestamp: int = None, synonyms: List[str] = None,
                is_derived: bool = False, elements: set[Element] = None):
        super().__init__(name, timestamp, synonyms, is_derived=is_derived, elements=elements)
        # A flag to prevent premature `_update_elements` calls during initialization
        self.__initializing = True
        self.types: set[Type] = types if types is not None else set()
        self.packages: set[Package] = packages if packages is not None else set()
        self.constraints: set[Constraint] = constraints if constraints is not None else set()
        self.associations: set[Association] = associations if associations is not None else set()
        self.generalizations: set[Generalization] = generalizations if generalizations is not None else set()
        # initialization is done, we update elements
        self.__initializing = False
        self._update_elements()

    def _update_elements(self):
        """Recalculates the elements property by combining all relevant sets."""
        if not self.__initializing:
            self.elements = set(self.types) | set(self.packages) | set(self.constraints) \
                            | set(self.associations) | set(self.generalizations)

    @property
    def types(self) -> set[Type]:
        """set[Type]: Get the set of types in the domain model."""
        return self.__types

    @types.setter
    def types(self, types: set[Type]):
        """
        set[Type]: Set the set of types in the domain model, including primitive data types.
        
        Raises:
            ValueError: if there are two types with the same name.
        """
        types = types | data_types
        names_seen = set()
        duplicates = set()

        for type_ in types:
            if type_.name in names_seen:
                duplicates.add(type_.name)
            names_seen.add(type_.name)

        if duplicates:
            raise ValueError(f"The model cannot have types with duplicate names: {', '.join(duplicates)}")
        self.__types = types
        self._update_elements()

    def get_type_by_name(self, type_name: str) -> Type:
        """Type: Gets an Type by name."""
        return next(
            (type_element for type_element in self.types if type_element.name == type_name), None
            )

    def add_type(self, type_: Type):
        """Type: Add a type (Class or DataType) to the set of types of the model."""
        self.types = self.types | {type_}

    @property
    def associations(self) -> set[Association]:
        """set[Association]: Get the set of associations in the domain model."""
        return self.__associations

    @associations.setter
    def associations(self, associations: set[Association]):
        """
        set[Association]: Set the set of associations in the domain model.
        
        Raises:
            ValueError: if there are two associations with the same name.
        """
        if associations is not None:
            names_seen = set()
            duplicates = set()

            for association in associations:
                if association.name in names_seen:
                    duplicates.add(association.name)
                names_seen.add(association.name)

            if duplicates:
                raise ValueError(f"The model cannot have associations with duplicate names: {', '.join(duplicates)}")

            self.__associations = associations
        else:
            self.__associations = set()

        self._update_elements()

    def add_association(self, association: Association):
        """Association: Add an association to the set of associations of the model."""
        self.associations = self.associations | {association}

    @property
    def generalizations(self) -> set[Generalization]:
        """set[Generalization]: Get the set of generalizations in the domain model."""
        return self.__generalizations

    @generalizations.setter
    def generalizations(self, generalizations: set[Generalization]):
        """set[Generalization]: Set the set of generalizations in the domain model."""
        if generalizations is not None:
            self.__generalizations = generalizations
        else:
            self.__generalizations = set()

        self._update_elements()

    def add_generalization(self, generalization: Generalization):
        """Generalization: Add a generalization to the set of generalizations of the model."""
        self.generalizations = self.generalizations | {generalization}

    def get_enumerations(self) -> set[Enumeration]:
        """set[Enumeration]: Get the set of enumerations in the domain model."""
        return {element for element in self.types if isinstance(element, Enumeration)}

    @property
    def packages(self) -> set[Package]:
        """set[Package]: Get the set of packages in the domain model."""
        return self.__packages

    @packages.setter
    def packages(self, packages: set[Package]):
        """
        set[Package]: Get the set of packages in the domain model.
        
        Raises:
            ValueError: if there are two packages with the same name.
        """
        if packages is not None:
            names_seen = set()
            duplicates = set()

            for package in packages:
                if package.name in names_seen:
                    duplicates.add(package.name)
                names_seen.add(package.name)

            if duplicates:
                raise ValueError(f"The model cannot have packages with duplicate names: {', '.join(duplicates)}")

            self.__packages = packages
        else:
            self.__packages = set()

        self._update_elements()

    @property
    def constraints(self) -> set[Constraint]:
        """set[Constraint]: Get the set of constraints in the domain model."""
        return self.__constraints

    @constraints.setter
    def constraints(self, constraints: set[Constraint]):
        """
        set[Constraint]: Get the set of constraints in the domain model.
        
        Raises:
            ValueError: if there are two constraints with the same name.
        """
        if constraints is not None:
            names = [constraint.name for constraint in constraints]
            if len(names) != len(set(names)):
                raise ValueError("The model cannot have two constraints with the same name")
            self.__constraints = constraints
        else:
            self.__constraints = set()

        self._update_elements()

    def get_classes(self) -> set[Class]:
        """set[Class]: Get all classes within the domain model."""
        return {element for element in self.types if isinstance(element, Class)}

    def get_class_by_name(self, class_name: str) -> Class:
        """Class: Gets a class by name."""
        return next(
            (element for element in self.types if isinstance(element, Class) and
             element.name == class_name), None
            )

    def classes_sorted_by_inheritance(self) -> list[Class]:
        """list[Class]: Get the list of classes ordered by inheritance."""
        from besser.utilities import sort_by_timestamp
        classes = sort_by_timestamp(self.get_classes())
        # Set up a dependency graph
        child_map = {cl: set() for cl in classes}
        # Populating the child_map based on generalizations (edges in top-sort graph)
        for cl in classes:
            for generalization in cl.generalizations:
                child_map[generalization.general].add(cl)
        # Helper function for DFS
        def dfs(cl, visited, sorted_list):
            visited.add(cl)
            for child in child_map[cl]:
                if child not in visited:
                    dfs(child, visited, sorted_list)
            sorted_list.append(cl)
        # Perform DFS from each node that hasn't been visited yet
        visited = set()
        sorted_list = []
        for cl in classes:
            if cl not in visited:
                dfs(cl, visited, sorted_list)
        sorted_list.reverse()
        return sorted_list

    def __repr__(self):
        return (
            f'DomainModel({self.name}, {self.types}, {self.associations}, {self.generalizations}, '
            f'{self.packages}, {self.constraints}, {self.timestamp}, {self.synonyms})'
            f'is_derived={self.is_derived}'
        )



from besser.BUML.notations.ocl.BOCLLexer import BOCLLexer
from besser.BUML.notations.ocl.BOCLParser import BOCLParser
from besser.BUML.notations.ocl.BOCLListener import BOCLListener
from besser.BUML.notations.ocl.RootHandler import Root_Handler
from antlr4 import *

class OCLParserWrapper:
    def __init__(self, dm, om):
        self.dm = dm
        self.om = om


    def parse(self, ocl):
        input_stream = InputStream(ocl.expression)
        rootHandler = Root_Handler(self.dm, self.om)
        rootHandler.set_context(ocl.context)
        lexer = BOCLLexer(input_stream)
        stream = CommonTokenStream(lexer)
        parser = BOCLParser(stream)
        tree = parser.oclFile()
        listener = BOCLListener(rootHandler)
        listener.preprocess(ocl.expression)
        walker = ParseTreeWalker()
        walker.walk(listener, tree)


        return True


"""
BUML code for cnn_rnn model.
https://github.com/ultimate010/crnn/blob/master/mr_cnn_rnn.py
"""
from besser.BUML.metamodel.nn import NN, Conv1D, PoolingLayer, \
    GRULayer, LinearLayer, EmbeddingLayer, TensorOp, DropoutLayer
from besser.generators.nn.pytorch.pytorch_code_generator import (
    PytorchGenerator
)
from besser.generators.nn.tf.tf_code_generator import TFGenerator


nn_model: NN = NN(name="my_model")
nn_model.add_layer(EmbeddingLayer(name="l1", actv_func=None,
                                  num_embeddings=5000, embedding_dim=50))
nn_model.add_layer(DropoutLayer(name="l2", rate=0.5))
nn_model.add_layer(Conv1D(name="l3", actv_func="relu", in_channels=50,
                          out_channels=200, kernel_dim=[4],
                          input_reused=True, name_module_input="l2"))
nn_model.add_layer(PoolingLayer(name="l4", actv_func=None, pooling_type="max",
                                dimension="1D", kernel_dim=[2]))
nn_model.add_layer(Conv1D(name="l5", actv_func="relu", in_channels=50,
                          out_channels=200, kernel_dim=[5], input_reused=True,
                          name_module_input="l2"))
nn_model.add_layer(PoolingLayer(name="l6", actv_func=None, pooling_type="max",
                                dimension="1D", kernel_dim=[2]))
nn_model.add_tensor_op(TensorOp(name="op1", tns_type="concatenate",
                                layers_of_tensors=["l4", "l6"],
                                concatenate_dim=-1))
nn_model.add_layer(DropoutLayer(name="l7", rate=0.15))
nn_model.add_layer(GRULayer(name="l8", actv_func=None, input_size=400,
                            hidden_size=100, batch_first=True,
                            return_type="last"))
nn_model.add_layer(LinearLayer(name="l9", actv_func="relu",
                               in_features=100, out_features=400))
nn_model.add_layer(DropoutLayer(name="l10", rate=0.10))
nn_model.add_layer(LinearLayer(name="l11", actv_func="sigmoid",
                               in_features=400, out_features=1))



pytorch_model = PytorchGenerator(model=nn_model, output_dir="output/cnn_rnn")
pytorch_model.generate()
tf_model = TFGenerator(model=nn_model, output_dir="output/cnn_rnn")
tf_model.generate()



"""
Module to test the NN metamodel
"""

import pytest

from besser.BUML.metamodel.nn import TensorOp, LinearLayer, EmbeddingLayer, \
    FlattenLayer, SimpleRNNLayer, Conv1D, Conv2D, Conv3D, PoolingLayer, \
    DropoutLayer, BatchNormLayer, LayerNormLayer, Dataset, Label, Image, \
    Structured, Configuration, NN


def run_tests():
    """Run tests"""
    pytest.main([__file__])

# Test tensorops
def test_tensorop():
    """Test tensorop"""
    l1: LinearLayer = LinearLayer(name="l1", actv_func="relu",
                                  in_features=10, out_features=20)
    l2: LinearLayer = LinearLayer(name="l2", actv_func="relu",
                                  in_features=10, out_features=20)
    reshape: TensorOp = TensorOp(name="reshape", tns_type="reshape",
                                 reshape_dim=[2, 2])
    permute: TensorOp = TensorOp(name="permute", tns_type="permute",
                                 permute_dim=[0, 2, 1])
    concatenate: TensorOp = TensorOp(name="concatenate",
                                     tns_type="concatenate",
                                     layers_of_tensors=[l1, l2],
                                     concatenate_dim=1)
    transpose: TensorOp = TensorOp(name="transpose", tns_type="transpose",
                                   transpose_dim=[0, 1])

    assert reshape.tns_type == 'reshape'
    assert reshape.reshape_dim == [2, 2]
    assert permute.permute_dim == [0, 2, 1]
    assert concatenate.layers_of_tensors == [l1, l2]
    assert transpose.transpose_dim == [0, 1]


# Test general layers
def test_general_layer():
    """Test general layer"""
    linear_layer: LinearLayer = LinearLayer(name="linear_layer",
                                            actv_func="relu",
                                            in_features=10, out_features=20,
                                            input_reused=True)
    flatten_layer: FlattenLayer = FlattenLayer(name="flatten_layer",
                                               actv_func="relu",
                                               start_dim=1, end_dim=-1)
    embedding_layer: EmbeddingLayer = EmbeddingLayer(
        name="embedding_layer", actv_func="relu", num_embeddings=10,
        embedding_dim=20, name_module_input="linear_layer"
    )

    assert linear_layer.name == "linear_layer"
    assert linear_layer.in_features == 10
    assert linear_layer.input_reused is True
    assert flatten_layer.end_dim == -1
    assert embedding_layer.num_embeddings == 10
    assert embedding_layer.actv_func == "relu"
    assert embedding_layer.name_module_input == "linear_layer"

# Test RNN layers
def test_rnn_layer():
    """Test rnn layer"""
    simple_rnn_layer: SimpleRNNLayer = SimpleRNNLayer(
        name="simple_rnn_layer", input_size=64, hidden_size=32,
        bidirectional=False, dropout=True, batch_first=True,
        return_type="last", input_reused=False
    )

    assert simple_rnn_layer.name == "simple_rnn_layer"
    assert simple_rnn_layer.input_size == 64
    assert simple_rnn_layer.hidden_size == 32
    assert simple_rnn_layer.return_type == "last"
    assert simple_rnn_layer.bidirectional is False
    assert simple_rnn_layer.dropout is True
    assert simple_rnn_layer.input_reused is False

#Test conv layers
def test_conv_layer():
    """Test conv layer"""
    conv1_layer: Conv1D = Conv1D(
        name="conv1_layer", actv_func="relu", in_channels=3, out_channels=6,
        kernel_dim=[3], stride_dim=[2]
    )
    conv2_layer: Conv2D = Conv2D(
        name="conv2_layer", actv_func="relu", in_channels=3, out_channels=6,
        kernel_dim=[3, 4], stride_dim=[2, 3]
    )
    conv3_layer: Conv3D = Conv3D(
        name="conv3_layer", actv_func="relu", in_channels=3, out_channels=6,
        kernel_dim=[3, 4, 5], stride_dim=[2, 3, 4]
    )

    assert conv1_layer.name == "conv1_layer"
    assert conv1_layer.in_channels == 3
    assert conv1_layer.kernel_dim[0] == 3
    assert conv1_layer.stride_dim[0] == 2
    assert conv2_layer.kernel_dim[1] == 4
    assert conv3_layer.stride_dim[2] == 4


#Test pooling layers
def test_pooling_layer():
    """Test pooling layer"""
    pool_avg1_layer: PoolingLayer = PoolingLayer(
        name="pool_avg1_layer", pooling_type="average", padding_type="same",
        stride_dim=[2], actv_func="relu", dimension="1D", kernel_dim=[3]
    )
    pool_adapt_avg_layer: PoolingLayer = PoolingLayer(
        name="pool_adapt_avg_layer", actv_func="relu", dimension="2D",
        pooling_type="adaptive_max", output_dim=[5, 7]
    )
    assert pool_avg1_layer.name == "pool_avg1_layer"
    assert pool_avg1_layer.dimension == "1D"
    assert pool_avg1_layer.kernel_dim[0] == 3
    assert pool_adapt_avg_layer.output_dim[1] == 7

#Test cnn layers padding
def test_cnn_padding():
    """Test cnn layers padding"""
    with pytest.raises(ValueError) as excinfo:
        conv1_layer: Conv1D = Conv1D(
            name="conv1_layer", actv_func="relu", padding_type="my_padding",
            out_channels=6, kernel_dim=[3], stride_dim=[2], in_channels=3
        )
    assert "Invalid padding type" in str(excinfo.value)

#Test pooling type
def test_pooling_type():
    """Test pooling type"""
    with pytest.raises(ValueError) as excinfo:
        pool_avg1_layer: PoolingLayer = PoolingLayer(
            name="pool_avg1_layer", actv_func="relu", kernel_dim=[3],
            stride_dim=[2], pooling_type="my_type", dimension="1D"
        )
    assert "Invalid pooling type" in str(excinfo.value)

#Test pooling dimension
def test_pooling_dimension():
    """Test pooling dimension"""
    with pytest.raises(ValueError) as excinfo:
        pool_avg1_layer: PoolingLayer = PoolingLayer(
            name="pool_avg1_layer", actv_func="relu", kernel_dim=[3],
            stride_dim=[2], pooling_type="average", dimension="my_dimension"
        )
    assert "Invalid pooling dimensionality" in str(excinfo.value)


#Test dropout layer
def test_dropout_layer():
    """Test dropout layer"""
    dropout_layer: DropoutLayer = DropoutLayer(name="dropout_layer", rate=0.5)
    assert dropout_layer.rate == 0.5


#Test normalization layer
def test_normalization_layer():
    """Test normalization layer"""
    batch_norm_layer: BatchNormLayer = BatchNormLayer(
        name="batch_norm_layer", actv_func="relu", num_features=3,
        dimension="3D"
    )
    layer_norm_layer: LayerNormLayer = LayerNormLayer(
        name="layer_norm_layer", actv_func="relu", normalized_shape=[3, 2, 4]
    )
    assert batch_norm_layer.num_features == 3
    assert layer_norm_layer.normalized_shape[0] == 3
    assert layer_norm_layer.normalized_shape[1] == 2
    assert layer_norm_layer.normalized_shape[2] == 4



#Test batch norm dimension
def test_batch_norm_dimension():
    """Test batch norm dimension"""
    with pytest.raises(ValueError) as excinfo:
        batch_norm_layer: BatchNormLayer = BatchNormLayer(
            name="batch_norm_layer", actv_func="relu",
            num_features=3, dimension="my_dimension"
        )
    assert "Invalid data dimensionality" in str(excinfo.value)


#Test activation function
def test_activation_function():
    """Test activation function"""
    conv1_layer: Conv1D = Conv1D(
        name="conv1_layer", actv_func="relu", in_channels=3,
        out_channels=6, kernel_dim=[3], stride_dim=[2]
    )
    assert conv1_layer.actv_func == "relu"


#Test dataset
def test_dataset():
    """Test dataset"""
    my_image = Image(shape=[32, 32, 32], normalize=True)
    training_data: Dataset = Dataset(
        name="training_data", path_data="c/my/path1", image=my_image,
        task_type="binary", input_format="csv"
    )
    test_data: Dataset = Dataset(name="test_data", path_data="c/my/path2")
    assert training_data.name == "training_data"
    assert training_data.path_data == "c/my/path1"
    assert training_data.input_format == "csv"
    assert training_data.task_type == "binary"
    assert training_data.image == my_image
    assert test_data.name == "test_data"
    assert test_data.path_data == "c/my/path2"


#Test data task type
def test_data_task_type():
    """Test data task type"""
    with pytest.raises(ValueError) as excinfo:
        training_data: Dataset = Dataset(
            name="training_data", path_data="c/my/path1",
            task_type="my_task", input_format="csv"
        )
    assert "Invalid value of task_type" in str(excinfo.value)

#Test label
def test_label():
    """Test label"""
    my_label: Label = Label(col_name="target", label_name="dog")
    assert my_label.col_name == "target"
    assert my_label.label_name == "dog"

#Test image feature
def test_image():
    """Test image feature"""
    my_image: Image = Image(shape=[125, 125, 125])
    assert my_image.shape[0] == 125
    assert my_image.shape[1] == 125
    assert my_image.shape[2] == 125

#Test structured feature
def test_structured():
    """Test structured feature"""
    my_feature: Structured = Structured(name="my_feature")
    assert my_feature.name == "my_feature"

#Test parameters
def test_configuration():
    """Test parameters"""
    my_configuration: Configuration = Configuration(
        batch_size=16, epochs=20, learning_rate=0.01, optimizer="sgd",
        loss_function="crossentropy", metrics=["f1-score", "accuracy"],
        weight_decay=0.001, momentum=0.1
    )

    assert my_configuration.batch_size == 16
    assert my_configuration.epochs == 20
    assert my_configuration.learning_rate == 0.01
    assert my_configuration.weight_decay == 0.001
    assert my_configuration.optimizer == "sgd"
    assert my_configuration.loss_function == "crossentropy"
    assert my_configuration.metrics[0] == "f1-score"
    assert my_configuration.momentum == 0.1
    assert len(my_configuration.metrics) == 2


#Test optimizer type
def test_optimizer_type():
    """Test optimizer type"""
    with pytest.raises(ValueError) as excinfo:
        my_configuration: Configuration = Configuration(
            batch_size=16, epochs=20, learning_rate=0.01,
            optimizer="my_optimizer", loss_function="crossentropy",
            metrics=["f1-score"], weight_decay=0.001
        )
    assert "Invalid value of optimizer" in str(excinfo.value)


#Test loss function type
def test_loss_function_type():
    """Test loss function type"""
    with pytest.raises(ValueError) as excinfo:
        my_configuration: Configuration = Configuration(
            batch_size=16, epochs=20, learning_rate=0.01, optimizer="sgd",
            loss_function="my_loss_function", metrics=["f1-score"],
            weight_decay=0.001
        )
    assert "Invalid value of loss_function" in str(excinfo.value)


#Test metrics
def test_metrics():
    """Test metrics"""
    with pytest.raises(ValueError) as excinfo:
        my_configuration: Configuration = Configuration(
            batch_size=16, epochs=20, learning_rate=0.01, optimizer="sgd",
            loss_function="crossentropy", metrics=["my_metric"],
            weight_decay=0.001
        )
    assert "Invalid metric(s) provided" in str(excinfo.value)



# Test nn model
def test_nn():
    """Test nn model"""
    nn_model: NN = NN(name="my_model")
    conv1_layer: Conv1D = Conv1D(
        name="conv1_layer", actv_func="relu", in_channels=3,
        out_channels=6, kernel_dim=[3], stride_dim=[2]
    )
    pool_avg1_layer: PoolingLayer = PoolingLayer(
        name="pool_avg1_layer", actv_func="relu", kernel_dim=[3],
        stride_dim=[2], pooling_type="average", dimension="1D",
        padding_type="same"
    )
    linear_layer: LinearLayer = LinearLayer(
        name="linear_layer", actv_func="relu", in_features=10, out_features=20
    )

    my_configuration: Configuration = Configuration(
        batch_size=16, epochs=20, learning_rate=0.01, optimizer="sgd",
        loss_function="crossentropy", metrics=["f1-score"],
        weight_decay=0.001
    )

    training_data: Dataset = Dataset(
        name="training_data", path_data="c/my/path1",
        task_type="binary", input_format="csv"
    )
    test_data: Dataset = Dataset(name="test_data", path_data="c/my/path2")

    nn_model.add_layer(conv1_layer)
    nn_model.add_layer(pool_avg1_layer)
    nn_model.add_layer(linear_layer)

    nn_model.add_configuration(my_configuration)
    nn_model.add_train_data(training_data)
    nn_model.add_test_data(test_data)


    assert nn_model.name == "my_model"
    assert len(nn_model.layers) == 3
    assert nn_model.layers[0].name == "conv1_layer"
    assert nn_model.configuration.epochs == 20
    assert nn_model.train_data == training_data
    assert nn_model.test_data == test_data

if __name__ == "__main__":
    run_tests()



from .deploymentParser import deploymentParser
from .deploymentListener import deploymentListener

class Deployment_BUML_Listener(deploymentListener):
    """
    Listener for parsing deployment-related models and writing the corresponding B-UML code.

    This class extends `deploymentListener` and overrides methods for handling different components
    of a deployment model.
    It writes the parsed deployment architecture into a specified output stream in the B-UML
    model format.

    Attributes:
        output (TextIO): The output stream where the B-UML model code is written.
        __cluster_list (list): A list to keep track of clusters in the deployment model.
    """

    def __init__(self, output):
        self.output = output
        self.__cluster_list: list = list()
        text = "from besser.BUML.metamodel.deployment import *\n\n# Deployment architecture model definition\n\n"
        self.output.write(text)

    def enterApplication(self, ctx: deploymentParser.ApplicationContext):
        name = ctx.ID().getText()
        repo = ctx.STRING(0).getText()
        port = ctx.INT(0).getText()
        cpu = ctx.INT(1).getText()
        memory = ctx.INT(2).getText()
        model = ctx.STRING(1).getText()
        text = name + " : Application = Application(name=\""+ name + "\", image_repo=" + repo + ", port=" + port + \
            ", required_resources=Resources(cpu=" + cpu + ", memory=" + memory + "), domain_model=" + model + ")\n"
        self.output.write(text)

    def enterService(self, ctx: deploymentParser.ServiceContext):
        name = ctx.ID(0).getText()
        port = ctx.INT(0).getText()
        target_port = ctx.INT(1).getText()
        protocol = ctx.protocol().getText().lower()
        type = ctx.service_type().getText()
        text = name + " : Service = Service(name=\""+ name + "\", port=" + port + ", target_port=" + target_port + \
            ", protocol=Protocol." + protocol + ", type=ServiceType." + type
        if (ctx.app is None):
            text += ")\n"
        else:
            text += ", application=" + ctx.ID(1).getText().replace(" ", "") + ")\n"
        self.output.write(text)

    def enterContainer(self, ctx: deploymentParser.ContainerContext):
        name = ctx.ID(0).getText()
        app = ctx.ID(1).getText()
        cpu = ctx.INT(0).getText()
        memory = ctx.INT(1).getText()
        text = name + " : Container = Container(name=\""+ name + "\", application=" + app + ", resources_limit=Resources(cpu=" + \
            cpu + ", memory=" + memory + "))\n"
        self.output.write(text)

    def enterDeployment(self, ctx: deploymentParser.DeploymentContext):
        name = ctx.ID(0).getText()
        replicas = ctx.INT().getText()
        container = ctx.ID(1).getText()
        text = name + " : Deployment = Deployment(name=\"" + name + "\", replicas=" + replicas + ", containers={" + \
            container.replace("\"", "").replace(" ", "")
        if len(ctx.ID()) > 2:
            for i in range(2, len(ctx.ID())):
                text += ", " + ctx.ID(i).getText()
        text += "})\n"
        self.output.write(text)

    def enterRegion(self, ctx: deploymentParser.RegionContext):
        name = ctx.ID_REG(0).getText()
        text = name.replace("-", "_") + " : Region = Region(name=\"" + name + "\", zones={})\n"
        self.output.write(text)

    def enterPublicCluster(self, ctx: deploymentParser.PublicClusterContext):
        text = "\n# Public cluster definition\n"
        name = ctx.ID(0).getText()
        nodes = ctx.INT().getText()
        provider = ctx.provider().getText()
        config_file = ctx.STRING().getText()
        text += name + " : PublicCluster = PublicCluster(name=\"" + name + "\", num_nodes=" + nodes + ", provider=Provider." + provider + \
            ", config_file=" + config_file
        self.output.write(text)
        self.__cluster_list.append(name)
    
    def enterService_list(self, ctx: deploymentParser.Service_listContext):
        service = ctx.ID(0).getText()
        text = ", services={" + service
        if len(ctx.ID()) > 1:
            for i in range(1, len(ctx.ID())):
                text += ", " + ctx.ID(i).getText()
        text += "}"
        self.output.write(text)

    def enterDeployment_list(self, ctx: deploymentParser.Deployment_listContext):
        deployment = ctx.ID(0).getText()
        text = ", deployments={" + deployment
        if len(ctx.ID()) > 1:
            for i in range(1, len(ctx.ID())):
                text += ", " + ctx.ID(i).getText()
        text += "}"
        self.output.write(text)        

    def enterRegion_list(self, ctx: deploymentParser.Region_listContext):
        region = ctx.ID_REG(0).getText().replace("-", "_")
        text = ", regions={" + region
        if len(ctx.ID_REG()) > 1:
            for i in range(1, len(ctx.ID_REG())):
                text += ", " + ctx.ID_REG(i).getText().replace("-", "_")
        text += "}"
        self.output.write(text)

    def exitPublicCluster(self, ctx: deploymentParser.PublicClusterContext):
        text = ")\n"
        self.output.write(text)

    def exitArchitecture(self, ctx: deploymentParser.ArchitectureContext):
        text = "\n# Deployment architecture model definition\n"
        text += "deployment_model : DeploymentModel = DeploymentModel(name=\"deployment_model\", clusters=" + list_to_str(self.__cluster_list) + ")\n"
        self.output.write(text)

def list_to_str(elements:list):
    """
        Method to transform a list of elements to string

        Args:
           elements (list): The list to transform.
    """
    if len(elements) == 0:
        str_list = "set()"
    else:
        str_list = ", ".join(elements)
        str_list = "{" + str_list + "}"
    return str_list



import os
from jinja2 import Environment, FileSystemLoader
from besser.BUML.metamodel.structural import DomainModel
from besser.generators import GeneratorInterface


class JavaGenerator(GeneratorInterface):

    def __init__(self, model: DomainModel, output_dir: str = None):
        super().__init__(model, output_dir)

    def generate(self):
        processed_associations = []
        for class_obj in self.model.classes_sorted_by_inheritance():
            file_path = self.build_generation_path(file_name=class_obj.name+".java")
            templates_path = os.path.join(os.path.dirname(
                os.path.abspath(__file__)), "templates")
            env = Environment(loader=FileSystemLoader(
                templates_path), trim_blocks=True, lstrip_blocks=True, extensions=['jinja2.ext.do'])
            template = env.get_template('java_template.py.j2')
            package_name = ""
            if self.output_dir is not None:
                if 'tmp' or 'AppData' in self.output_dir:
                    package_name = None
                else:
                    package_name = self.output_dir
            else:
                package_name = None
            with open(file_path, mode="w") as f:
                generated_code = template.render(class_obj=class_obj,
                                                 processed_associations=processed_associations,
                                                 package_name=package_name)
                f.write(generated_code)
                print("Code generated in the location: " + file_path)



"""
This module defines the `PyTorchGenerator` class that generates
PyTorch code for neural networks based on the B-UML model.
"""

from typing import Callable
from besser.BUML.metamodel.nn import NN
from besser.generators.nn.pytorch.utils_pytorch import SetupLayerSyntax, \
    get_tensorop_syntax
from besser.generators.nn.nn_code_generator import NNCodeGenerator


class PytorchGenerator(NNCodeGenerator):
    """
    PytorchGenerator is a class that inherits from `NNCodeGenerator`.
    It generates Pytorch code for neural networks training and evaluation 
    based on the B-UML input model.

    Args:
        model (NN): An instance of the NN Model class representing 
            the B-UML model.
        setup_layer (SetupLayerSyntax): The class that defines
            the syntax of layers.
        setup_tensorop (Callable): The function that defines the
            syntax of tensorops.
        output_dir (str, optional): The output directory where the 
            generated code will be saved. Defaults to None.
        file_name (str): The name of the file where the generated
            code is stored.
        template (str): The name of the jinja template.
    """
    def __init__(self, model: NN, output_dir: str,
                 generation_type: str = "subclassing",
                 channel_last: bool = True):

        setup_layer: SetupLayerSyntax = SetupLayerSyntax
        setup_tensorop: Callable = get_tensorop_syntax

        template_dir: str = "pytorch"
        file_name: str = "pytorch_nn.py"

        super().__init__(model, setup_layer, setup_tensorop, generation_type,
                         channel_last, template_dir, file_name, output_dir)



import os
from jinja2 import Environment, FileSystemLoader
from besser.BUML.metamodel.structural import DomainModel
from besser.generators import GeneratorInterface
from besser.utilities import sort_by_timestamp

class PythonGenerator(GeneratorInterface):
    """
    PythonGenerator is a class that implements the GeneratorInterface and is responsible
    for generating the Python domain model code based on the input B-UML model.

    Args:
        model (DomainModel): An instance of the DomainModel class representing the B-UML model.
        output_dir (str, optional): The output directory where the generated code will be 
            saved. Defaults to None.
    """
    def __init__(self, model: DomainModel, output_dir: str = None):
        super().__init__(model, output_dir)

    def generate(self, *args):
        """
        Generates Python domain model code based on the provided B-UML model and saves it to 
        the specified output directory.
        If the output directory was not specified, the code generated will be stored in the 
        <current directory>/output folder.

        Returns:
            None, but store the generated code as a file named classes.py 
        """
        file_path = self.build_generation_path(file_name="classes.py")
        templates_path = os.path.join(os.path.dirname(
            os.path.abspath(__file__)), "templates")
        env = Environment(loader=FileSystemLoader(templates_path))
        template = env.get_template('python_classes_template.py.j2')
        with open(file_path, mode="w", encoding='utf-8') as f:
            generated_code = template.render(domain=self.model, sort_by_timestamp=sort_by_timestamp)
            f.write(generated_code)
            print("Code generated in the location: " + file_path)



import os
import subprocess
import sys
from jinja2 import Environment, FileSystemLoader
from besser.BUML.metamodel.structural import DomainModel
from besser.generators import GeneratorInterface
from besser.utilities import sort_by_timestamp

class DjangoGenerator(GeneratorInterface):
    """
    DjangoGenerator is a class that implements the GeneratorInterface and is responsible for generating
    the Django application code based on the input B-UML model.

    Args:
        model (DomainModel): An instance of the DomainModel class representing the B-UML model.
        project_name (str): The name of the Django project.
        app_name (str): The name of the Django app.
        containerization (bool, optional): Flag indicating if containerization support should be added 
            (False as defaults).
        output_dir (str, optional): The output directory where the generated code will be saved. Defaults to None.
    """

    def __init__(self, model: DomainModel, project_name: str, app_name: str,
                containerization: bool = False, output_dir: str = None):
        super().__init__(model, output_dir)
        self.project_name: str = project_name
        self.app_name: str = app_name
        self.containerization: bool = containerization
        templates_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates")
        self.env = Environment(loader=FileSystemLoader(templates_path),
                                trim_blocks=True,
                                lstrip_blocks=True,
                                extensions=['jinja2.ext.do'])

    def generate(self, *args):
        """
        Generates Django application code based on the provided B-UML model and saves it to the specified 
        output directory. If the output directory was not specified, the code generated will be stored in 
        the <current directory>/<project_name>.

        Returns:
            None, but store the generated code.
        """

        # Build Django project
        subprocess.run(['django-admin', 'startproject', self.project_name], check=True)
        # Create Django app
        subprocess.run([sys.executable, 'manage.py', 'startapp', self.app_name],
                        cwd=self.project_name, check=True)
        # Update settings.py
        self.update_settings()
        # Create requirements.txt
        self.create_file_from_template('requirements.txt.j2', 'requirements.txt')
        # Create docker files if containerization is True
        if self.containerization is True:
            self.create_file_from_template('docker_compose.j2', 'docker-compose.yml')
            self.create_file_from_template('dockerfile.j2', 'Dockerfile')
            self.create_file_from_template('entrypoint.sh.j2', 'entrypoint.sh')
        # Create models.py
        self.create_file_from_template('models.py.j2', os.path.join(self.app_name, "models.py"))
        # Create admin.py
        self.create_file_from_template('admin.py.j2', os.path.join(self.app_name, "admin.py"))

    def create_file_from_template(self, template_name, output_name):
        """Create a file from a Jinja2 template."""
        template = self.env.get_template(template_name)
        file_path = os.path.join(self.project_name, output_name)
        with open(file_path, mode="w", newline='\n', encoding='utf-8') as f:
            f.write(template.render(app_name=self.app_name,
                                    project_name=self.project_name,
                                    model=self.model,
                                    sort=sort_by_timestamp))

    def update_settings(self):
        """Update the configuration in settings.py."""
        settings_file_path = os.path.join(self.project_name, self.project_name, 'settings.py')
        new_database_config = ""
        if self.containerization is True:
            new_database_config = """
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql',
        'NAME': os.environ.get('POSTGRES_NAME'),
        'USER': os.environ.get('POSTGRES_USER'),
        'PASSWORD': os.environ.get('POSTGRES_PASSWORD'),
        'HOST': 'db',
        'PORT': 5432,
    }
}
"""
        jazzmin_settings = f"""
# Admin template configuration
JAZZMIN_SETTINGS = {{
    # title of the window (Will default to current_admin_site.site_title if absent or None)
    "site_title": "{self.project_name} - Admin",

    # Title on the login screen (19 chars max) (defaults to current_admin_site.site_header if absent or None)
    "site_header": "{self.project_name}",

    # Title on the brand (19 chars max) (defaults to current_admin_site.site_header if absent or None)
    "site_brand": "{self.project_name}",

    # Logo to use for your site, must be present in static files, used for brand on top left
    # "site_logo": "img/logo.png",

    # Links to put along the top menu
    "topmenu_links": [

        # Url that gets reversed (Permissions can be added)
        {{"name": "Home",  "url": "admin:index", "permissions": ["auth.view_user"]}},

        # App with dropdown menu to all its models pages (Permissions checked against models)
        {{"app": "{self.app_name}"}},

        # model admin to link to (Permissions checked against model)
        {{"model": "auth.User"}},
    ],
}}
"""
        try:
            with open(settings_file_path, 'r', encoding='utf-8') as file:
                content = file.readlines()

            # Ensure 'import os' is present
            if not any(line.startswith('import os') for line in content):
                for index, line in enumerate(content):
                    if line.strip() and not line.strip().startswith('#'):
                        content.insert(index, 'import os\n')
                        break

            if self.containerization is True:
                # Replace the DATABASES section
                start_index, end_index = None, None
                for index, line in enumerate(content):
                    if 'DATABASES' in line and '=' in line:
                        start_index = index
                    if start_index is not None and line.strip() == '}':
                        end_index = index
                        break

                if start_index is not None and end_index is not None:
                    content = content[:start_index] + [new_database_config] + content[end_index + 2:]

            # Add the app to INSTALLED_APPS
            for index, line in enumerate(content):
                if line.strip().startswith('INSTALLED_APPS') and '=' in line:
                    # Find the start of the list
                    open_bracket_index = index
                    while '[' not in content[open_bracket_index]:
                        open_bracket_index += 1

                    # Find the end of the list
                    close_bracket_index = open_bracket_index
                    while ']' not in content[close_bracket_index]:
                        close_bracket_index += 1

                    # Add the app if not already in the list
                    apps_section = content[open_bracket_index:close_bracket_index + 1]
                    if f"'{self.app_name}'," not in ''.join(apps_section):
                        # Insert the app just before the closing bracket
                        content.insert(close_bracket_index, f"    '{self.app_name}',\n")
                    if f"'{'jazzmin'}'," not in ''.join(apps_section):
                        # Insert the jazzmin app just before the closing bracket
                        content.insert(open_bracket_index + 1, "    'jazzmin',\n")
                    break

            # Add the JAZZMIN_SETTINGS block at the end of the file
            if jazzmin_settings.strip() not in ''.join(content):
                content.append(f"\n{jazzmin_settings}\n")

            # Write the updated settings back to the file
            with open(settings_file_path, 'w', encoding='utf-8') as file:
                file.writelines(content)

        except (IOError, OSError) as e:
            print(f"An I/O error occurred: {e}")
        except ValueError as e:
            print(f"A value error occurred: {e}")



from besser.BUML.metamodel.structural import NamedElement, Class, Property, Model
from enum import Enum


# FileSourceType
class FileSourceType:
    """Represents the type of a file source.

    Args:
        name (str): The name of the file source type.
        type (str): The type of the file source, such as 'FileSystem', 'LocalStorage', or 'DatabaseFileSystem'.

    Attributes:
        name (str): The name of the file source type.
        type (str): The type of the file source, such as 'FileSystem', 'LocalStorage', or 'DatabaseFileSystem'.
    """
    
    def __init__(self, name: str, type: str):
        self.type: str = type
        self.name: str = name

    @property
    def name(self) -> str:
        """str: Get the name of the file source type."""
        return self.__name

    @name.setter
    def name(self, name: str):
        """str: Set the name of the file source type."""
        self.__name = name

    @property
    def type(self) -> str:
        """str: Get the type of the file source."""
        return self.__type

    @type.setter
    def type(self, type: str):
        """
        str: Set the type of the file source.

        Raises:
            ValueError: If the type provided is not one of the allowed values: 'FileSystem', 'LocalStorage',
                or 'DatabaseFileSystem'.
        """
        if type not in ['FileSystem', 'LocalStorage', 'DatabaseFileSystem']:
            raise ValueError("Invalid value of type")
            self.__type = type

    def __repr__(self):
        return f'FileSourceType({self.name}, type={self.type})'

# CollectionSourceType
class CollectionSourceType:
    """Represents the type of a collection source.

    Args:
        name (str): The name of the collection source type.
        type (str): The type of the collection source, such as 'List', 'Table', 'Tree', 'Grid', 'Array', or 'Stack'.

    Attributes:
        name (str): The name of the collection source type.
        type (str): The type of the collection source, such as 'List', 'Table', 'Tree', 'Grid', 'Array', or 'Stack'.
    """
    
    def __init__(self, name: str, type: str):
        self.type: str = type
        self.name: str = name

    @property
    def name(self) -> str:
        """str: Get the name of the collection source type."""
        return self.__name

    @name.setter
    def name(self, name: str):
        """str: Set the name of the collection source type."""
        self.__name = name

    @property
    def type(self) -> str:
        """str: Get the type of the collection source."""
        return self.__type

    @type.setter
    def type(self, type: str):
        """str: Set the type of the collection source.

        Raises:
            ValueError: If the type provided is not one of the allowed values: 'List', 'Table', 'Tree', 'Grid',
                'Array', or 'Stack'.
        """
        if type not in ['List', 'Table', 'Tree', 'Grid', 'Array', 'Stack']:
            raise ValueError("Invalid value of type")
            self.__type = type

    def __repr__(self):
        return f'CollectionSourceType({self.name}, type={self.type})'

class ButtonActionType(Enum):
    """Represents a button action type.
    """
    Add = "Add"
    ShowList = "Show List"
    OpenForm = "Open Form"
    SubmitForm = "Submit Form"
    Cancel = "Cancel"
    Save = "Save"
    Delete = "Delete"
    Confirm = "Confirm"
    Navigate = "Navigate"
    Search = "Search"
    Filter = "Filter"
    Sort = "Sort"
    Send = "Send"
    Share = "Share"
    Settings = "Settings"
    Back = "Back"
    Next = "Next"
    View = "View"
    Select = "Select"
    Login = "Login"
    Logout = "Sign Out"
    Help = "Help"
    About= "About"
    Exit = "Exit"



class ButtonType(Enum):
    """Represents a button type.
    """
    RaisedButton = "Raised Button"
    TextButton = "Text Button"
    OutlinedButton = "Outlined Button"
    IconButton = "Icon Button"
    FloatingActionButton = "FloatingActionButton"
    DropdownButton = "Dropdown Button"
    ToggleButtons = "Toggle Buttons"
    iOSStyleButton = "iOS-style Button"
    CustomizableButton = "Customizable Button"


#DataSource
class DataSource:
    """Represents a data source.

    Args:
        name (str): The name of the data source.

    Attributes:
        name (str): The name of the data source.
    """

    def __init__(self, name: str):
        self.name: str = name

    @property
    def name(self) -> str:
      """str: Get the name of the data source."""
      return self.__name

    @name.setter
    def name(self, name: str):
      """str: Set the name of the data source."""
      self.__name = name

    def __repr__(self):
      return f'DataSource({self.name})'

#ModelElementDataSource
class ModelElement(DataSource):
    """Represents a data source associated with a model element.

    Args:
        name (str): The name of the model element data source.
        dataSourceClass (Class): The class representing the data source.
        fields: set[Property]: The fields representing the attributes of the model element.

    Attributes:
        name (str): The name of the model element data source.
        dataSourceClass (Class): The class representing the data source.
        fields: set[Property]: The fields representing the attributes of the model element.
    """
    
    def __init__(self, name: str, dataSourceClass: Class, fields: set[Property]):
        super().__init__(name)
        self.dataSourceClass: Class = dataSourceClass
        self.fields: set[Property]= fields

    @property
    def dataSourceClass(self) -> Class:
        """Class: Get the class representing the data source."""
        return self.__dataSourceClass

    @dataSourceClass.setter
    def dataSourceClass(self, dataSourceClass: Class):
        """Class: Set the class representing the data source."""
        self.__dataSourceClass = dataSourceClass

    @property
    def fields(self) -> set[Property]:
        """set[Property]: Get the set of properties (fields) of the model element."""
        return self.__fields

    @fields.setter
    def fields(self, fields: set[Property]):
        """set[Property]: Set the set of properties (fields) of the model element."""
        if fields is not None:
            names = [field.name for field in fields]
            if len(names) != len(set(names)):
                raise ValueError("A model element cannot have two fields with the same name.")
        self.__fields = fields

    def __repr__(self):
      return f'ModelElement({self.name}, {self.dataSourceClass},{self.fields})'

#FileDataSource
class File(DataSource):
    """Represents a data source that is a file.

    Args:
        name (str): The name of the file data source.
        type (FileSourceType): The type of the file data source.

    Attributes:
        name (str): The name of the file data source.
        type (FileSourceType): The type of the file data source.
    """
    
    def __init__(self, name: str, type:FileSourceType):
        super().__init__(name)
        self.type: FileSourceType = type

    @property
    def type(self) -> FileSourceType:
      """FileSourceType: Get the type of the file data source."""
      return self.__type

    @type.setter
    def type(self, type: FileSourceType):
      """FileSourceType: Set the type of the file data source."""
      self.__type = type

    def __repr__(self):
      return f'File({self.name}, {self.type})'


#CollectionDataSource
class Collection(DataSource):
    """Represents a data source that is a collection.

    Args:
        name (str): The name of the collection data source.
        type (CollectionSourceType): The type of the collection data source.

    Attributes:
        name (str): The name of the collection data source.
        type (CollectionSourceType): The type of the collection data source.
    """
    def __init__(self, name: str, type:CollectionSourceType):
        super().__init__(name)
        self.type: CollectionSourceType = type

    @property
    def type(self) -> CollectionSourceType:
      """CollectionSourceType: Get the type of the collection data source."""
      return self.__type

    @type.setter
    def type(self, type: CollectionSourceType):
      """CollectionSourceType: Set the type of the collection data source."""
      self.__type = type

    def __repr__(self):
      return f'Collection({self.name}, {self.type})'

#ViewElement
class ViewElement(NamedElement):
    """Represents a view element.

    Args:
        name (str): The name of the view element.
        description (str): the description of the view element.

    Attributes:
        name (str): The name of the view element.
        description (str): the description of the view element.
    """
    
    def __init__(self, name: str, description: str, visibility: str = "public"):
        super().__init__(name, visibility)
        self.description: str = description

    @property
    def description(self) -> str:
        """str: Get the description of the view element."""
        return self.__description

    @description.setter
    def description(self, description: str):
        """str: Set the description of the view element."""
        self.__description = description

    def __repr__(self):
        return f'ViewElement({self.name})'

#ViewComponent
class ViewComponent(ViewElement):
    """Represents a view component.

    Args:
        name (str): The name of the view component.
        description (str): The description of the view component.

    Attributes:
        name (str): The name of the view component.
        description (str): The description of the view component.
    """
    
    def __init__(self, name: str, description: str):
        super().__init__(name, description)

    def __repr__(self):
        return f'ViewComponent({self.name}, description={self.description})'

#ViewContainer
class ViewContainer(ViewElement):
    """Represents a view container.

    Args:
        name (str): The name of the view container.
        description (str): The description of the view container.

    Attributes:
        name (str): The name of the view container.
        description (str): The description of the view container.
    """
    
    def __init__(self, name: str, description: str, view_elements: set[ViewElement]):
        super().__init__(name, description)
        self.view_elements: set[ViewElement] = view_elements

    @property
    def view_elements(self) -> set[ViewElement]:
        """set[ViewComponent]: Get the set of view elements on the screen."""
        return self.__view_elements

    @view_elements.setter
    def view_elements(self, view_elements: set[ViewElement]):
       """set[ViewElement]: Set the set of view elements on the screen."""
       if view_elements is not None:
            names = [view_element.name for view_element in view_elements]
            if len(names) != len(set(names)):
                raise ValueError("A screen cannot have two elements with the same name.")
       self.__view_elements = view_elements

    def __repr__(self):
        return f'ViewContainer({self.name}, description={self.description}, view_elements={self.view_elements})'

#Screen
class Screen(ViewContainer):
    """Represents a screen.

    Args:
        name (str): The name of the screen.
        view_elements (set[ViewElement]): The set of view elements on the screen.
        x_dpi (str): The X DPI (dots per inch) of the screen.
        y_dpi (str): The Y DPI (dots per inch) of the screen.
        size (str): The size of the screen.

    Attributes:
        name (str): The name of the screen.
        view_elements (set[ViewElement]): The set of view elements on the screen.
        x_dpi (str): The X DPI (dots per inch) of the screen.
        y_dpi (str): The Y DPI (dots per inch) of the screen.
        size (str): The size of the screen.
    """
    
    def __init__(self, name: str, description: str, view_elements: set[ViewElement], x_dpi: str, y_dpi: str, size: str):
        super().__init__(name, description, view_elements)
        self.x_dpi: str = x_dpi
        self.y_dpi: str = y_dpi
        self.size: str = size
    
    @property
    def x_dpi(self) -> str:
      """str: Get the X DPI (dots per inch) of the screen."""
      return self.__x_pdi

    @x_dpi.setter
    def x_dpi(self, x_dpi: str):
      """str: Set the X DPI (dots per inch) of the screen."""
      self.__x_pdi = x_dpi

    @property
    def y_dpi(self) -> str:
      """str: Get the Y DPI (dots per inch) of the screen."""
      return self.__y_pdi

    @y_dpi.setter
    def y_dpi(self, y_dpi: str):
     """str: Set the Y DPI (dots per inch) of the screen."""
     self.__y_pdi = y_dpi

    @property
    def size(self) -> str:
      """str: Get the size of the screen."""
      return self.__size

    @size.setter
    def size(self, size: str):
        """str: Set the size of the screen.

        Raises:
            ValueError: If the size provided is not one of the allowed options: 'SmallScreen','MediumScreen', 'LargScreen', 'xLargeScreen'
        """

        if size not in ['SmallScreen', 'MediumScreen', 'LargScreen', 'xLargeScreen']:
           raise ValueError("Invalid value of size")
           self.__size = size

    def __repr__(self):
      return f'Screen({self.name}, {self.x_dpi}, {self.y_dpi}, {self.size}, {self.view_elements})'

#Module
class Module(NamedElement):
    """Represents a module.

    Args:
        name (str): name (str): The name of the module.
        screens (set[Screen]): The set of screens contained in the module.

    Attributes:
        name (str): name (str): The name of the module.
        screens (set[Screen]): The set of screens contained in the module.
    """
    
    def __init__(self, name: str, screens: set[Screen], visibility: str = "public"):
        super().__init__(name, visibility)
        self.screens: set[Screen] = screens

    @property
    def screens(self) -> set[Screen]:
        """set[Screen]: Get the set of screens contained."""
        return self.__screens

    @screens.setter
    def screens(self, screens: set[Screen]):
       """set[Screen]: Set the set of screens contained."""
       if screens is not None:
            names = [screen.name for screen in screens]
            if len(names) != len(set(names)):
                raise ValueError("A module cannot have two screens with the same name.")
       self.__screens = screens

    def __repr__(self):
        return f'Module({self.name}, {self.screens})'

# DataList is a type of ViewComponent
class DataList(ViewComponent):
    """Represents a list component that encapsulates properties unique to lists, such as list sources.

    Args:
        name (str): The name of the list.
        list_sources (set[DataSource]): The set of data sources associated with the list.

    Attributes:
        name (str): The name of the list.
        list_sources (set[DataSource]): The set of data sources associated with the list.
    """
    
    def __init__(self, name: str, description: str, list_sources: set[DataSource]):
        super().__init__(name, description)
        self.list_sources: set[DataSource] = list_sources
    
    @property
    def list_sources(self) -> set[DataSource]:
        """set[DataSource]: Get the set of data sources associated with the list."""
        return self.__list_sources

    @list_sources.setter
    def list_sources(self, list_sources: set[DataSource]):
        """set[DataSource]: Set the set of data sources associated with the list."""
        if list_sources is not None:
            names = [DataSource.name for DataSource in list_sources]
            if len(names) != len(set(names)):
               raise ValueError("A list cannot have two items with the same name.")
        self.__list_sources = list_sources   

    def __repr__(self):
     return f'DataList({self.name}, {self.list_sources})'

# Button is a type of ViewComponent
class Button(ViewComponent):
    """Represents a button component and encapsulates specific properties of a button, such as its name and label.

    Args:
        name (str): The name of the button.
        description (str): The description of the button.
        label (str): The label of the button.
        buttonType (ButtonType): The type of the button.
        actionType (ButtonActionType): The action type of the button.
        targetScreen (Screen, optional): The target Screen associated with the button when the actionType is "Navigate".

    Attributes:
        name (str): The name of the button.
        description (str): The description of the button.
        label (str): The label of the button.
        buttonType (ButtonType): The type of the button.
        actionType (ButtonActionType): The action type of the button.
        targetScreen (Screen, optional): The target Screen associated with the button when the actionType is "Navigate"
    """
    
    def __init__(self, name: str, description: str, label: str, buttonType: ButtonType, actionType: ButtonActionType, targetScreen: Screen = None):
        super().__init__(name, description)
        self.label=label
        self.buttonType: ButtonType = buttonType
        self.actionType: ButtonActionType=actionType
        self.targetScreen: Screen = targetScreen

    @property
    def label(self) -> str:
        """str: Get the label of the button."""
        return self.__label

    @label.setter
    def label(self, label: str):
        """str: Set the label of the button."""
        self.__label = label
    
    @property
    def buttonType(self) -> ButtonType:
        """str: Get the type of the button."""
        return self.__buttonType

    @buttonType.setter
    def buttonType(self, buttonType: ButtonType):
        """str: Set the type of the button."""
        self.__buttonType = buttonType

    @property
    def actionType(self) -> ButtonActionType:
        """str: Get the action type of the button."""
        return self.__actionType

    @actionType.setter
    def actionType(self, actionType: ButtonActionType):
        """str: Set the action type of the button."""
        self.__actionType = actionType

    @property
    def targetScreen(self) -> Screen:
        """Type: Get the target Screen of the button."""
        return self.__targetScreen

    @targetScreen.setter
    def targetScreen(self, targetScreen: Screen):
        """
        Set the target Screen associated with the button.

        Args:
            targetScreen (Screen): The target Screen to be associated with the button.

        Raises:
            ValueError: If the actionType is 'Navigate' and the target Screen is not an instance of the Screen class.
            ValueError: If the actionType is not 'Navigate' and an target Screen is specified.

        """
        if self.actionType == ButtonActionType.Navigate:
            if targetScreen is None:
                raise ValueError("A target Screen must be specified for the button when the actionType is 'Navigate'.")
            elif not isinstance(targetScreen, Screen):
                raise ValueError("The target Screen must be an instance of the Screen class when the actionType is 'Navigate'.")
        elif targetScreen is not None:
            raise ValueError("A target Screen cannot be specified for the button when the actionType is not 'Navigate'.")
        self.__targetScreen = targetScreen

    def __repr__(self):
     return f'Button({self.name},{self.label}, {self.description}, {self.buttonType}, {self.actionType})'

# Image is a type of ViewComponent
class Image(ViewComponent):
    """Represents an image component and encapsulates the specific properties of a image, such as its name.

    Args:
        name (str): The name of the image.
        
    Attributes:
        name (str): The name of the image.
    """
    
    def __init__(self, name: str, description: str):
        super().__init__(name, description)

    def __repr__(self):
     return f'Image({self.name},{self.description})'


# InputField is a type of ViewComponent
class InputField(ViewComponent):
    """Represents an input field component and encapsulates specific properties of an input field, such as its type and validation rules.

     Args:
        name (str): The name of the input field.
        description (str): The description of the input field.
        fieldType (str): The type of the input field.
        validationRules (str): The validation rules for the input field.

    Attributes:
        name (str): The name of the input field.
        description (str): The description of the input field.
        fieldType (str): The type of the input field.
        validationRules (str): The validation rules for the input field.
    """
    
    def __init__(self, name: str, description: str, fieldType: str, validationRules: str):
        super().__init__(name, description)
        self.fieldType: str= fieldType
        self.validationRules: str = validationRules

    @property
    def fieldType(self) -> str:
      """str: Get the type of the field."""
      return self.__fieldType


    @fieldType.setter
    def fieldType(self, fieldType: str):
        """str: Set the type of the field.

        Raises:
            ValueError: If the type provided is not one of the allowed options: 'Text','Number', 'Email', 'Password', 'Date', 'Time', 'File', 'Color', 'Range', 'URL', 'Tel', and 'Search'
        """

        if fieldType not in ['Text','Number', 'Email', 'Password', 'Date', 'Time', 'File', 'Color', 'Range', 'URL', 'Tel', 'Search']:
           raise ValueError("Invalid value of fieldType")
           self.__fieldType = fieldType

    @property
    def validationRules(self) -> str:
        """str: Get the validation rules of the input field."""
        return self.__validationRules

    
    @validationRules.setter
    def validationRules(self, str):
        """str: Set the validation rules of the input field."""
        self.__validationRules = validationRules

    def __repr__(self):
     return f'InputField({self.name},{self.description}, {self.fieldType}, {self.validationRules})'

# Form is a type of ViewComponent
class Form(ViewComponent):
    """Represents a form component and encapsulates the specific properties of a form, such as its name.

    Args:
        name (str): The name of the form.
        description (str): The description of the form.
        inputFields (set[InputField]): The set of input fields contained in the form.
        
    Attributes:
        name (str): The name of the form.
        description (str): The description of the form.
        inputFields (set[InputField]): The set of input fields contained in the form.
    """
    
    def __init__(self, name: str, description: str, inputFields: set[InputField]):
        super().__init__(name, description)
        self.inputFields: set[InputField] = inputFields
     
    @property
    def inputFields(self) -> set[InputField]:
        """set[InputField]: Get the set of input Fields contained in the form."""
        return self.__inputFields

    @inputFields.setter
    def inputFields(self, inputFields: set[InputField]):
       """set[InputField]: Set the set of input Fields contained in the form."""
       self.__inputFields = inputFields

    def __repr__(self):
     return f'Form({self.name},{self.description}, {self.inputFields})'

# MenuItem
class MenuItem:
    """Represents an item of a menu.

    Args:
        label (str): The label of the menu item.
        
    Attributes:
        label (str): The label of the menu item.
    """
    
    def __init__(self, label: str):
        self.label: str = label

    def __repr__(self):
     return f'MenuItem({self.label})'

# Menu is a type of ViewComponent
class Menu(ViewComponent):
    """Represents a menu component and encapsulates the specific properties of a menu, such as its name.

    Args:
        name (str): The name of the menu.
        description (str): The description of the menu.
        menuItems (set[MenuItem]): The set of menu items contained in the menu.
        
    Attributes:
        name (str): The name of the menu.
        description (str): The description of the menu.
        menuItems (set[MenuItem]): The set of menu items contained in the menu.
    """
    
    def __init__(self, name: str, description: str, menuItems: set[MenuItem]):
        super().__init__(name, description)
        self.menuItems: set[MenuItem] = menuItems

    @property
    def menuItems(self) -> set[MenuItem]:
        """set[MenuItem]: Get the set of menuItems."""
        return self.__menuItems

    @menuItems.setter
    def menuItems(self, menuItems: set[MenuItem]):
        """set[MenuItem]: Set the set of menuItems."""
        self.__menuItems = menuItems

    def __repr__(self):
     return f'Menu({self.name},{self.description}, {self.menuItems})'

#Application
class Application(Model):
    """It is a subclass of the NamedElement class and encapsulates the properties and behavior of an application, including its name, 
       package, version code, version name, modules, description, and screen compatibility.

    Args:
        name (str): The name of the application.
        package (str): The package of the application.
        versionCode (str): The version code of the application.
        versionName (str): The version name of the application.
        modules (set[Module]): The set of modules contained in the application.
        description (str): The description of the application.
        screenCompatibility (bool): Indicates whether the application has screen compatibility.

    Attributes:
        name (str): The name of the application.
        package (str): The package of the application.
        versionCode (str): The version code of the application.
        versionName (str): The version name of the application.
        modules (set[Module]): The set of modules contained in the application.
        description (str): The description of the application.
        screenCompatibility (bool): Indicates whether the application has screen compatibility.
    """
    def __init__(self, name: str, package: str, versionCode: str, versionName: str, modules: set[Module], description: str, screenCompatibility: bool = False):
        super().__init__(name)
        self.package: str = package
        self.versionCode: str = versionCode
        self.versionName: str = versionName
        self.description: str = description
        self.modules: set[Module] = modules
        self.screenCompatibility: str = screenCompatibility

    @property
    def package(self) -> str:
        """str: Get the package of the application."""
        return self.__package

    @package.setter
    def package(self, package: str):
        """str: Set the package of the application."""
        self.__package = package

    @property
    def versionCode(self) -> str:
        """str: Get the version code of the application."""
        return self.__versionCode

    @versionCode.setter
    def versionCode(self, versionCode: str):
        """str: Set the version code of the application."""
        self.__versionCode = versionCode

    @property
    def versionName(self) -> str:
        """str: Get the version name of the application."""
        return self.__versionName

    @versionName.setter
    def versionName(self, versionName: str):
        """str: Set the version name of the application."""
        self.__versionName = versionName

    @property
    def description(self) -> str:
        """str: Get the description of the application."""
        return self.__description

    @description.setter
    def description(self, description: str):
        """str: Set the description of the application."""
        self.__description = description

    @property
    def screenCompatibility(self) -> bool:
        """bool: Get the screen compatibility of the application."""
        return self.__screenCompatibility

    @screenCompatibility.setter
    def screenCompatibility(self, screenCompatibility: bool):
        """bool: Set the screen compatibility of the application."""
        self.__screenCompatibility = screenCompatibility

    @property
    def modules(self) -> set[Module]:
        """set[Module]: Get the set of modules contained in the application."""
        return self.__modules

    @modules.setter
    def modules(self, modules: set[Module]):
       """set[Module]: Set the set of modules contained in the application."""
       if modules is not None:
            names = [module.name for module in modules]
            if len(names) != len(set(names)):
                raise ValueError("An app cannot have two modules with the same name")
       self.__modules = modules


    def __repr__(self):
        return f'Application({self.name}, {self.package}, {self.versionCode}, {self.versionName},{self.description},{self.screenCompatibility}, {self.modules})'








"""
This module defines the `NNCodeGenerator` class that is inherited
by `TFGenerator` and `PytorchGenerator` to generates code for neural 
networks based on the B-UML model.
"""

import os
from typing import Dict, Callable, Union
from jinja2 import Environment, FileSystemLoader
from besser.BUML.metamodel.nn import NN
from besser.generators import GeneratorInterface
from besser.generators.nn.tf.utils_tf import (
    SetupLayerSyntax as SetupLayerTF
)
from besser.generators.nn.pytorch.utils_pytorch import (
    SetupLayerSyntax as SetupLayerTorch
)
from besser.generators.nn.pytorch.utils_pytorch import adjust_actv_func_name
from besser.generators.nn.utils_nn import handle_layer, handle_tensorop, \
    add_in_out_var_to_subnn


class NNCodeGenerator(GeneratorInterface):
    """
    NNCodeGenerator is a class that implements the GeneratorInterface and 
    is inherited by `TFGenerator` and `PytorchGenerator` to generates code 
    for neural networks training and evaluation based on the B-UML model. 

    Args:
        model (NN): An instance of the NN Model class representing 
            the B-UML model.
        setup_layer (Union[SetupLayerTF, SetupLayerTorch]): The class 
            that defines the syntax of layers.
        get_tensorop_syntax (Callable): The function that defines the
            syntax of tensorops.
        output_dir (str, optional): The output directory where the 
            generated code will be saved. Defaults to None.
        file_name (str): The name of the file where the generated
            code is stored.
        template_name (str): The name of the jinja template.
        template_dir (str): The name of the directory where the jinja 
            template `template_name` is stored. Either `tf` or `pytorch`.
    """
    def __init__(self, model: NN,
                 setup_layer: Union[SetupLayerTF, SetupLayerTorch],
                 get_tensorop_syntax: Callable, generation_type: str,
                 channel_last: bool, template_dir: str,
                 file_name: str = "nn.py", output_dir: str = None):

        super().__init__(model, output_dir)
        self.setup_layer: Union[SetupLayerTF, SetupLayerTorch] = setup_layer
        self.get_tensorop_syntax: Callable = get_tensorop_syntax
        self.generation_type: str = generation_type
        self.channel_last: bool = channel_last
        self.template_dir: str = template_dir
        self.file_name: str = file_name

        if self.generation_type == "subclassing":
            self.template_name = f"template_{template_dir}_subclassing.py.j2"
        else:
            self.template_name = f"template_{template_dir}_sequential.py.j2"

        self.modules_details: Dict = self.get_modules_details()


    def get_modules_details(self) -> str:
        """
        A module can be a layer, a sub_nn or a tensorop.
        The `modules_details` dict is created to keep track of 
        the syntax of modules, their tensor input variables, and 
        their tensor output variables in the forward method. 
        It has this structure:
                {"name_module": [syntax, out_var, in_var]}
        - syntax: The syntax of calling the module.
        - out_var: the output tensor variable of the module.
        - in_var: the input tensor variable of module.
        Example (TensoFlow):
                  {"l2": 
                  ["self.l2 = layers.Dense(units=40, activation='relu')",
                  "x_1", 
                  "x"]}

        For the case of layers, an additional element is added to 
        the list, representing the layer object.
        """
        counter_subnn = 0
        modules_details = {}
        if "torch" in self.template_name:
            actv_func = True
        else:
            actv_func = False

        is_seq = False
        if self.generation_type == "sequential":
            is_seq = True
        for module in self.model.modules:
            module_type = module.__class__.__name__
            if module_type == "NN":
                subnn_details = {}
                for sub_nn_layer in module.layers:
                    subnn_details = handle_layer(
                        sub_nn_layer, self.setup_layer, subnn_details,
                        actv_func, is_seq, self.channel_last, is_subnn=True
                    )
                name_sub_nn = f"{module.name}_{counter_subnn}_nn"
                modules_details[name_sub_nn] = subnn_details
                counter_subnn += 1
                modules_details = add_in_out_var_to_subnn(modules_details)
            elif module_type != "TensorOp":
                modules_details = handle_layer(
                    module, self.setup_layer, modules_details,
                    actv_func, is_seq, self.channel_last, is_subnn=False
                )
            else:
                modules_details = handle_tensorop(
                    module, modules_details, self.get_tensorop_syntax)
        if actv_func:
            modules_details = adjust_actv_func_name(modules_details)
        return modules_details

    def generate(self, *args):
        """
        Generates NN code based on the provided B-UML model and saves 
        it to the specified output directory.
        If the output directory was not specified, the code generated 
        will be stored in the <current directory>/output folder.

        Returns:
            None, but stores the generated code as a file named nn_code.py
        """
        file_name = f"{self.file_name[:-3]}_{self.generation_type}.py"
        file_path = self.build_generation_path(file_name=file_name)
        templates_path = os.path.join(os.path.dirname(
            os.path.abspath(__file__)), self.template_dir, "templates")
        env = Environment(loader=FileSystemLoader(templates_path))
        template = env.get_template(self.template_name)

        with open(file_path, mode="w", encoding="utf-8") as f:
            generated_code = template.render(
                model=self.model, modules_details=self.modules_details,
                generation_type=self.generation_type)
            f.write(generated_code)
            print("Code generated in the location: " + file_path)



import warnings
from besser.BUML.metamodel.structural import DomainModel, Class, Multiplicity, Property, \
    BinaryAssociation, Generalization, Enumeration, EnumerationLiteral, Method, Parameter
from .PlantUMLParser import PlantUMLParser
from .PlantUMLListener import PlantUMLListener

class BUMLGenerationListener(PlantUMLListener):
    """
       This listener class generates a B-UML structural model from a parse-tree that 
       representing a plantUML textual model
    """

    VISIBILITY = {"+": "public", "-": "private", "#": "protected", "~": "package"}

    def __init__(self):
        self.__buml_model = None

    def get_buml_model(self):
        """DomainModel: Retrieves the B-UML model instance."""
        return self.__buml_model

    def enterDomainModel(self, ctx: PlantUMLParser.DomainModelContext):
        self.__buml_model = DomainModel(name="DomainModel")
        classes = self.find_descendant_nodes_by_type(node=ctx,
                                                target_type=PlantUMLParser.ClassContext)
        for cl in classes:
            abstract = True if cl.abstract() else False
            new_class: Class = Class(name=cl.ID().getText(), is_abstract=abstract)
            self.__buml_model.add_type(new_class)
            # check extends
            if cl.extends():
                general_cl_name = cl.extends().ID().getText()
                general_class = self.__buml_model.get_class_by_name(class_name=general_cl_name)
                if general_class is None:
                    general_class = Class(name=general_cl_name)
                self.__buml_model.add_type(general_class)
                # create generalization
                new_generalization: Generalization = Generalization(general=general_class,
                                                                    specific=new_class)
                self.__buml_model.add_generalization(new_generalization)

        enums = self.find_descendant_nodes_by_type(node=ctx,
                                                target_type=PlantUMLParser.EnumerationContext)
        for enum in enums:
            new_enum: Enumeration = Enumeration(name=enum.ID().getText())
            self.__buml_model.add_type(new_enum)

    def find_descendant_nodes_by_type(self, node, target_type):
        """
        Recursively finds and returns all descendant nodes of a specified type.

        Args:
            node: The node to search for descendants. This can be any node in the parse tree.
            target_type: The type of node to match against. This should be a class type that 
                        the nodes are expected to be instances of.

        Returns:
            A list of nodes that are instances of the specified target type. 
            If no matching nodes are found, an empty list is returned.
        """
        matching_nodes = []

        if isinstance(node, target_type):
            matching_nodes.append(node)

        for i in range(node.getChildCount()):
            child = node.getChild(i)
            matching_nodes.extend(self.find_descendant_nodes_by_type(child, target_type))

        return matching_nodes

    def enterAttribute(self, ctx: PlantUMLParser.AttributeContext):
        cl_name = ctx.parentCtx.ID().getText()
        attr_name = ctx.ID().getText()

        visibility = self.VISIBILITY[ctx.visibility().getText()] if ctx.visibility() \
            else "public"

        if ctx.dType().primitiveData():
            primitive_type = "str" if ctx.dType().primitiveData().getText() == "string" \
                else ctx.dType().primitiveData().getText()
            attr_type = self.__buml_model.get_type_by_name(primitive_type)
        else:
            attr_type = self.__buml_model.get_type_by_name(ctx.dType().ID().getText())
        if attr_type is None:
            raise ValueError("Use a valid type for the \"" + attr_name +  "\" attribute")

        new_attr: Property = Property(name=attr_name, type=attr_type, visibility=visibility)
        cl = self.__buml_model.get_class_by_name(class_name=cl_name)
        cl.add_attribute(attribute=new_attr)


    def enterMethod(self, ctx: PlantUMLParser.MethodContext):
        cl_name = ctx.parentCtx.ID().getText()
        method_name = ctx.ID().getText()

        visibility = self.VISIBILITY[ctx.visibility().getText()] if ctx.visibility() \
            else "public"
        abstract = False
        if ctx.modifier():
            abstract = True if ctx.modifier().getText() == "{abstract}" else False

        method_type = None
        if ctx.dType():
            if ctx.dType().primitiveData():
                primitive_type = "str" if ctx.dType().primitiveData().getText() == "string" \
                    else ctx.dType().primitiveData().getText()
                method_type = self.__buml_model.get_type_by_name(primitive_type)
            else:
                method_type = self.__buml_model.get_type_by_name(ctx.dType().ID().getText())
                if method_type is None:
                    raise ValueError("Use a valid type for the \"" + method_name +  "\" method")

        new_method: Method = Method(name=method_name,
                                    visibility=visibility,
                                    is_abstract=abstract,
                                    type=method_type)

        cl = self.__buml_model.get_class_by_name(class_name=cl_name)
        cl.add_method(method=new_method)

    def enterParameter(self, ctx: PlantUMLParser.ParameterContext):
        cl_name = ctx.parentCtx.parentCtx.ID().getText()
        method_name = ctx.parentCtx.ID().getText()
        param_name = ctx.ID().getText()

        if ctx.dType().primitiveData():
            primitive_type = "str" if ctx.dType().primitiveData().getText() == "string" \
                else ctx.dType().primitiveData().getText()
            param_type = self.__buml_model.get_type_by_name(primitive_type)
        else:
            param_type = self.__buml_model.get_type_by_name(ctx.dType().ID().getText())
        if param_type is None:
            raise ValueError("Use a valid type for the \"" + param_name +  "\" parameter")

        default_value = None
        if ctx.value():
            if ctx.value().ID():
                default_value = ctx.value().ID().getText()
            elif ctx.value().INT():
                default_value = int(ctx.value().INT().getText())
            elif ctx.value().FLOAT():
                default_value = float(ctx.value().FLOAT().getText())
            if ctx.value().D_QUOTE(0) and ctx.value().D_QUOTE(1):
                default_value = str(default_value)

        new_param: Parameter = Parameter(name=param_name,
                                        type= param_type,
                                        default_value=default_value)
        methods = self.__buml_model.get_class_by_name(class_name=cl_name).methods
        for method in methods:
            if method.name == method_name:
                method.add_parameter(new_param)

    def enterInheritance(self, ctx: PlantUMLParser.InheritanceContext):
        if ctx.inh_left:
            general = self.__buml_model.get_class_by_name(ctx.ID(0).getText())
            specific = self.__buml_model.get_class_by_name(ctx.ID(1).getText())
        else:
            general = self.__buml_model.get_class_by_name(ctx.ID(1).getText())
            specific = self.__buml_model.get_class_by_name(ctx.ID(0).getText())
        new_generalization: Generalization = Generalization(general=general, specific=specific)
        self.__buml_model.add_generalization(new_generalization)

    def enterAssociation(self, ctx: PlantUMLParser.AssociationContext):
        cl_left = self.__buml_model.get_class_by_name(ctx.ID(0).getText())
        cl_right = self.__buml_model.get_class_by_name(ctx.ID(1).getText())
        navigation = [True, True]
        mult_left: Multiplicity = Multiplicity(min_multiplicity=1, max_multiplicity=1)
        mult_right: Multiplicity = Multiplicity(min_multiplicity=1, max_multiplicity=1)
        composition = [False, False]

        if ctx.ID(2) is None:
            assoc_name = cl_left.name + "_" + cl_right.name
            warnings.warn(
                f"No name was provided for the association between '{cl_left.name}' and "
                f"'{cl_right.name}'. A default name '{assoc_name}' will be auto-assigned.",
                UserWarning
            )
        else:
            assoc_name = ctx.ID(2).getText()
            for type_ in self.__buml_model.types:
                if type_.name.lower() == assoc_name.lower():
                    warnings.warn(
                        f"The association name '{assoc_name}' is identical or similar to "
                        f"the class/enumeration/type '{type_.name}', which may cause issues "
                        f"with some code generators.", UserWarning
                    )

        if ctx.unidirectional():
            unidirectional_ctx = ctx.unidirectional()
            navigation[0] = bool(unidirectional_ctx.nav_l)
            navigation[1] = bool(unidirectional_ctx.nav_r)

        if ctx.composition():
            composition_ctx = ctx.composition()
            composition[0] = bool(composition_ctx.comp_l)
            composition[1] = bool(composition_ctx.comp_r)

        if ctx.c_left:
            mult_left = self.get_cardinality(ctx.c_left)
        if ctx.c_right:
            mult_right = self.get_cardinality(ctx.c_right)

        end_left: Property = Property(name=ctx.ID(0).getText(),
                                    type=cl_left,
                                    multiplicity=mult_left,
                                    is_composite=composition[0],
                                    is_navigable=navigation[0])
        end_right: Property = Property(name=ctx.ID(1).getText(),
                                    type=cl_right,
                                    multiplicity=mult_right,
                                    is_composite=composition[1],
                                    is_navigable=navigation[1])
        new_association:BinaryAssociation = BinaryAssociation(name=assoc_name,
                                    ends={end_left, end_right})

        self.__buml_model.add_association(new_association)

    def get_cardinality(self, ctx: PlantUMLParser.CardinalityContext):
        """
        Extracts and constructs the cardinality from the given context.

        Args:
            ctx (PlantUMLParser.CardinalityContext): The context object that 
                contains the cardinality values to be parsed.

        Returns:
            Multiplicity: An B-UML Multiplicity object representing the minimum 
            and maximum cardinality derived from the context.
        """
        if ctx.cardinalityVal(0).INT():
            min_cardinality = max_cardinality = int(ctx.cardinalityVal(0).INT().getText())
        else:
            min_cardinality = 0
            max_cardinality = "*"

        if ctx.cardinalityVal(1):
            if ctx.cardinalityVal(1).INT():
                max_cardinality = int(ctx.cardinalityVal(1).INT().getText())
            else:
                max_cardinality = "*"

        multiplicity: Multiplicity = Multiplicity(min_multiplicity=min_cardinality,
                                                  max_multiplicity=max_cardinality)
        return multiplicity

    def enterEnumLiteral(self, ctx: PlantUMLParser.EnumLiteralContext):
        new_literal: EnumerationLiteral = EnumerationLiteral(name=ctx.ID().getText())
        enum = self.__buml_model.get_type_by_name(ctx.parentCtx.ID().getText())
        enum.add_literal(literal=new_literal)
